# -*- coding: utf-8 -*-
# wasp_general/io.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import io
import time
import gzip
import bz2

from wasp_general.verify import verify_type, verify_subclass, verify_value
from wasp_general.crypto.aes import WAES
from wasp_general.crypto.hash import WHash


class WHashIO:

	def __init__(self, hash_name):
		self.__hash_name = hash_name
		self.__hash_obj = WHash.generator(hash_name).new(b'')

	@verify_type(b=bytes)
	def update_hash(self, b):
		self.__hash_obj.update(b)

	def hash_name(self):
		return self.__hash_name

	def hexdigest(self):
		return self.__hash_obj.hexdigest()


class WResponsiveIO:

	class IOTerminated(Exception):
		pass

	def __init__(self, stop_event):
		self.__stop_event = stop_event

	def stop_event(self):
		return self.__stop_event


class WIOCounter:

	def __init__(self):
		self.__bytes_processed = 0
		self.__start_at = None
		self.__finished_at = None

	def bytes_processed(self):
		return self.__bytes_processed

	def start_counter(self):
		if self.__start_at is None:
			self.__start_at = time.time()
		else:
			raise RuntimeError('Unable to start counter two times in a row')

	def stop_counter(self):
		if self.__finished_at is None:
			self.__finished_at = time.time()

	def rate(self):
		start_at = self.__start_at
		if start_at is None:
			raise RuntimeError('Unable to calculate rate without start method called')

		finished_at = self.__finished_at
		if finished_at is None:
			finished_at = time.time()

		return self.bytes_processed() / (finished_at - start_at)

	@verify_type('paranoid', processed_bytes=int)
	def __iadd__(self, processed_bytes):
		self.increase_counter(processed_bytes)
		return self

	@verify_type(processed_bytes=int)
	def increase_counter(self, processed_bytes):
		self.__bytes_processed += processed_bytes

	def reset(self):
		self.__bytes_processed = 0
		self.__start_at = None
		self.__finished_at = None


class WThrottlingIO(WIOCounter):

	__default_maximum_timeout__ = 1.5

	@verify_type(throttling_to=(int, float, None), maximum_timeout=(int, float, None))
	@verify_value(maximum_timeout=lambda x: x is None or x > 0)
	def __init__(self, throttling_to=None, maximum_timeout=None):
		WIOCounter.__init__(self)
		self.__throttling_to = throttling_to
		self.__maximum_timeout = \
			maximum_timeout if maximum_timeout is not None else self.__default_maximum_timeout__

	def throttling_to(self):
		return self.__throttling_to

	def maximum_timeout(self):
		return self.__maximum_timeout

	def check_rate(self):
		current_rate = self.rate()
		max_rate = self.throttling_to()
		if max_rate is not None and current_rate > max_rate:
			rate_delta = current_rate - max_rate
			sleep_time = self.bytes_processed() / rate_delta
			time.sleep(min(sleep_time, self.maximum_timeout()))


class WIOChainLink:

	def __init__(self, io_cls, *args, **kwargs):
		self.__io_cls = io_cls
		self.__args = args
		self.__kwargs = kwargs

	def io_cls(self):
		return self.__io_cls

	def io_obj(self, raw):
		return self.io_cls()(raw, *self.__args, **self.__kwargs)


class WIOChain:

	@verify_type(links=WIOChainLink)
	def __init__(self, last_io_obj, *links):
		self.__chain = [last_io_obj]

		for link in links:
			next_io_obj = link.io_obj(last_io_obj)
			self.__chain.append(next_io_obj)
			last_io_obj = next_io_obj

	def first_io(self):
		return self.__chain[-1]

	def instance(self, io_cls):
		for io_obj in self:
			if isinstance(io_obj, io_cls) is True:
				return io_obj

	def __iter__(self):
		chain = self.__chain.copy()
		chain.reverse()
		for io_obj in chain:
			yield io_obj


class WAESWriter(io.BufferedWriter):
	""" File-like writer with transparent encryption
	"""

	@verify_type(cipher=WAES)
	def __init__(self, raw, cipher):
		""" Create new encryption writer

		:param cipher: cipher to use. As written data size may differ - cipher must be constructed with
		padding object
		:param raw: target file-like object to write to
		"""
		io.BufferedWriter.__init__(self, raw)

		self.__cipher_padding = cipher.mode().padding()
		if self.__cipher_padding is None:
			raise ValueError('AES cipher must be created with "padding" option')

		self.__cipher = cipher.cipher()
		self.__cipher_block_size = cipher.mode().key_size()
		self.__buffer = b''

	@verify_type(b=(bytes, memoryview))
	def write(self, b):
		""" Encrypt and write data

		:param b: data to encrypt and write

		:return: None
		"""
		self.__buffer += bytes(b)
		bytes_written = 0
		while len(self.__buffer) >= self.__cipher_block_size:
			io.BufferedWriter.write(self, self.__cipher.encrypt(self.__buffer[:self.__cipher_block_size]))
			self.__buffer = self.__buffer[self.__cipher_block_size:]
			bytes_written += self.__cipher_block_size
		return len(b)

	def flush(self):
		if len(self.__buffer) > 0:
			data = self.__cipher_padding.pad(self.__buffer, self.__cipher_block_size)
			encrypted_data = self.__cipher.encrypt(data)
			io.BufferedWriter.write(self, encrypted_data)
		self.__buffer = b''
		io.BufferedWriter.flush(self)


class WHashCalculationWriter(io.BufferedWriter, WHashIO):

	@verify_type(hash_name=str)
	def __init__(self, raw, hash_name):
		io.BufferedWriter.__init__(self, raw)
		WHashIO.__init__(self, hash_name)

	@verify_type('paranoid', b=(bytes, memoryview))
	def write(self, b):
		self.update_hash(bytes(b))
		io.BufferedWriter.write(self, b)
		return len(b)


class WThrottlingWriter(io.BufferedWriter, WThrottlingIO):

	@verify_type('paranoid', throttling_to=(int, float, None), maximum_timeout=(int, float, None))
	@verify_value('paranoid', maximum_timeout=lambda x: x is None or x > 0)
	def __init__(self, raw, throttling_to=None, maximum_timeout=None):
		io.BufferedWriter.__init__(self, raw)
		WThrottlingIO.__init__(self, throttling_to=throttling_to, maximum_timeout=maximum_timeout)
		self.start_counter()

	def close(self, *args, **kwargs):
		self.stop_counter()
		io.BufferedWriter.close(self, *args, **kwargs)

	@verify_type(b=(bytes, memoryview))
	def write(self, b):
		self.check_rate()
		io.BufferedWriter.write(self, b)
		data_length = len(b)
		self.increase_counter(data_length)
		return data_length


class WResponsiveWriter(io.BufferedWriter, WResponsiveIO):

	def __init__(self, raw, stop_event):
		io.BufferedWriter.__init__(self, raw)
		WResponsiveIO.__init__(self, stop_event)

	@verify_type(b=(bytes, memoryview))
	def write(self, b):
		if self.stop_event().is_set():
			raise WResponsiveIO.IOTerminated('Stop event was set')
		io.BufferedWriter.write(self, b)
		return len(b)


class WDiscardWriterResult(io.BufferedWriter):

	@verify_type(b=(bytes, memoryview))
	def write(self, b):
		return len(b)


class WWriterChainLink(WIOChainLink):

	@verify_subclass(writer_cls=io.BufferedWriter)
	def __init__(self, writer_cls, *args, **kwargs):
		WIOChainLink.__init__(self, writer_cls, *args, **kwargs)


class WWriterChain(WIOChain, io.BufferedWriter):

	@verify_type(links=WWriterChainLink)
	def __init__(self, last_io_obj, *links):
		WIOChain.__init__(self, last_io_obj, *links)
		io.BufferedWriter.__init__(self, self.first_io())

	def flush(self):
		io.BufferedWriter.flush(self)
		for link in self:
			link.flush()

	def close(self):
		for link in self:
			link.close()
		io.BufferedWriter.close(self)


class WBufferedIOReader(io.BufferedReader):

	def __init__(self, raw):
		io.BufferedReader.__init__(self, raw)

	def writable(self, *args, **kwargs):
		return False

	def read(self, size=-1):
		if size == 0:
			return self.read_chunk(size)

		result_buffer = self.create_buffer()
		chunk_size = io.DEFAULT_BUFFER_SIZE
		read_chunk = True
		while read_chunk is True:
			next_chunk = self.read_chunk(chunk_size)

			if size > 0:
				if chunk_size >= size:
					read_chunk = False
				size -= chunk_size
			elif next_chunk == b'':
				read_chunk = False

			result_buffer = self.append_buffer(result_buffer, next_chunk)

		return bytes(result_buffer)

	def read_chunk(self, size):
		return self.raw.read(size)

	@classmethod
	def create_buffer(cls):
		return b''

	@classmethod
	def append_buffer(cls, buffer, data):
		buffer += data
		return buffer


class WDiscardReaderResult(WBufferedIOReader):

	@classmethod
	def append_buffer(cls, buffer, data):
		return buffer


class WGzipReader(WBufferedIOReader):

	def __init__(self, raw):
		WBufferedIOReader.__init__(self, raw)
		self.__gzip = gzip.GzipFile(fileobj=raw)

	def read_chunk(self, size):
		return self.__gzip.read(size=size)

	def close(self, *args, **kwargs):
		self.__gzip.close()
		WBufferedIOReader.close(self)


class WBzip2Reader(WBufferedIOReader):

	def __init__(self, raw):
		WBufferedIOReader.__init__(self, raw)
		self.__bzip2 = bz2.BZ2File(raw)

	def read_chunk(self, size):
		return self.__bzip2.read(size=size)

	def close(self, *args, **kwargs):
		self.__bzip2.close()
		WBufferedIOReader.close(self)


class WHashCalculationReader(WBufferedIOReader, WHashIO):

	@verify_type(hash_name=str)
	def __init__(self, raw, hash_name):
		WBufferedIOReader.__init__(self, raw)
		WHashIO.__init__(self, hash_name)

	def read_chunk(self, size):
		result = WBufferedIOReader.read_chunk(self, size)
		self.update_hash(result)
		return result


class WResponsiveReader(WBufferedIOReader, WResponsiveIO):

	def __init__(self, raw, stop_event):
		WBufferedIOReader.__init__(self, raw)
		WResponsiveIO.__init__(self, stop_event)

	def read_chunk(self, size):
		if self.stop_event().is_set():
			raise WResponsiveIO.IOTerminated('Stop event was set')
		return WBufferedIOReader.read_chunk(self, size)


class WThrottlingReader(WBufferedIOReader, WThrottlingIO):

	@verify_type('paranoid', throttling_to=(int, float, None), maximum_timeout=(int, float, None))
	@verify_value('paranoid', maximum_timeout=lambda x: x is None or x > 0)
	def __init__(self, raw, throttling_to=None, maximum_timeout=None):
		WBufferedIOReader.__init__(self, raw)
		WThrottlingIO.__init__(self, throttling_to=throttling_to, maximum_timeout=maximum_timeout)
		self.start_counter()

	def close(self, *args, **kwargs):
		self.stop_counter()
		WBufferedIOReader.close(self, *args, **kwargs)

	def read_chunk(self, size):
		self.check_rate()
		result = WBufferedIOReader.read_chunk(self, size)
		self.increase_counter(len(result))
		return result


class WReaderChainLink(WIOChainLink):

	@verify_subclass(reader_cls=io.BufferedReader)
	def __init__(self, reader_cls, *args, **kwargs):
		WIOChainLink.__init__(self, reader_cls, *args, **kwargs)


class WReaderChain(WIOChain, io.BufferedReader):

	@verify_type(links=WReaderChainLink)
	def __init__(self, last_io_obj, *links):
		WIOChain.__init__(self, last_io_obj, *links)
		io.BufferedReader.__init__(self, self.first_io())

	def close(self):
		for link in self:
			link.close()
		io.BufferedReader.close(self)



# -*- coding: utf-8 -*-
# wasp_general/os/linux/mounts.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import os
import subprocess

from wasp_general.verify import verify_type, verify_value


class WMountPoint:
	""" This class supply information about system mount points. It allows to find mount point by a custom file path.
	"""

	__mounts_file__ = '/proc/mounts'
	""" System file with current mounts information
	"""

	@verify_type(mount_record=str)
	def __init__(self, mount_record):
		""" Create new mount point information by parsing mount point record (a single line from /proc/mounts
		file)

		:param mount_record: line to parse
		"""
		parsed_record = mount_record.split()
		if len(parsed_record) != 6:
			raise RuntimeError('Unable to parse mount record: %s' % mount_record)

		device = parsed_record[0]
		self.__device = os.path.realpath(device) if os.path.isabs(device) else device
		self.__device_name = os.path.basename(self.__device)
		self.__path = parsed_record[1]
		self.__fs = parsed_record[2]
		self.__options = tuple(parsed_record[3].split(','))

		def check_flag(flag):
			if flag == '0':
				return False
			elif flag == '1':
				return True
			else:
				raise RuntimeError('Invalid dump-flag spotted')

		self.__dump_flag = check_flag(parsed_record[4])
		self.__fsck_order = int(parsed_record[5])

	def device(self):
		""" Return a mount point device path

		:return: str
		"""
		return self.__device

	def device_name(self):
		""" Return a mount point device name (a base name of a device path)

		:return: str
		"""
		return self.__device_name

	def path(self):
		""" Return a mount point path (path where this mount point is mounted to)

		:return: str
		"""
		return self.__path

	def fs(self):
		""" Return a mount point filesystem name

		:return: str
		"""
		return self.__fs

	def options(self):
		""" Return a mount point options (options with which this point was mounted)

		:return: tuple of str
		"""
		return self.__options

	def dump_flag(self):
		""" Return dump flag for this mount point

		:return: bool
		"""
		return self.__dump_flag

	def fsck_order(self):
		""" Return fsck (filesystem check) order for this mount point

		:return: int
		"""
		return self.__fsck_order

	@classmethod
	def mounts(cls):
		""" Return tuple of current mount points

		:return: tuple of WMountPoint
		"""
		result = []
		with open(cls.__mounts_file__) as f:
			for mount_record in f:
				result.append(WMountPoint(mount_record))
		return tuple(result)

	@classmethod
	@verify_type(file_path=str)
	@verify_value(file_path=lambda x: len(x) > 0)
	def mount_point(cls, file_path):
		""" Return mount point that, where the given path is reside on

		:param file_path: target path to search
		:return: WMountPoint or None (if file path is outside current mount points)
		"""
		mount = None

		for mp in cls.mounts():
			mp_path = mp.path()
			if file_path.startswith(mp_path) is True:
				if mount is None or len(mount.path()) <= len(mp_path):
					mount = mp

		return mount

	@classmethod
	@verify_type(device=str, mount_directory=str, fs=(str, None), options=(list, tuple, set, None))
	@verify_type(cmd_timeout=(int, float, None), sudo=bool)
	@verify_value(device=lambda x: len(x) > 0, mount_directory=lambda x: len(x) > 0)
	@verify_value(fs=lambda x: x is None or len(x) > 0, cmd_timeout=lambda x: x is None or x > 0)
	def mount(cls, device, mount_directory, fs=None, options=None, cmd_timeout=None, sudo=False):
		""" Mount a device to mount directory

		:param device: device to mount
		:param mount_directory: target directory where the given device will be mounted to
		:param fs: optional, filesystem on the specified device. If specifies - overrides OS filesystem \
		detection with this value.
		:param options: specifies mount options (OS/filesystem dependent)
		:param cmd_timeout: if specified - timeout with which this mount command should be evaluated (if \
		command isn't complete within the given timeout - an exception will be raised)
		:param sudo: whether to use sudo to run mount command

		:return: None
		"""
		cmd = [] if sudo is False else ['sudo']
		cmd.extend(['mount', device, os.path.abspath(mount_directory)])
		if fs is not None:
			cmd.extend(['-t', fs])
		if options is not None and len(options) > 0:
			cmd.append('-o')
			cmd.extend(options)

		subprocess.check_output(cmd, timeout=cmd_timeout)

	@classmethod
	@verify_type(device_or_directory=str, cmd_timeout=(int, float, None), sudo=bool)
	@verify_value(device_or_directory=lambda x: len(x) > 0, cmd_timeout=lambda x: x is None or x > 0)
	def umount(cls, device_or_directory, cmd_timeout=None, sudo=False):
		""" Unmount device (or mount directory)

		:param device_or_directory: device name or mount directory to unmount
		:param cmd_timeout: if specified - timeout with which this unmount command should be evaluated (if \
		command isn't complete within the given timeout - an exception will be raised)
		:param sudo: whether to use sudo to run mount command

		:return: None
		"""
		cmd = [] if sudo is False else ['sudo']
		cmd.extend(['umount', device_or_directory])
		subprocess.check_output(cmd, timeout=cmd_timeout)



# -*- coding: utf-8 -*-
# wasp_general/network/service_v2/proto.py
#
# Copyright (C) 2018 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod
from enum import Enum

from wasp_general.verify import verify_type, verify_subclass
from wasp_general.uri import WURI, WSchemeHandler, WSchemeCollection
from wasp_general.config import WConfigSelection


# noinspection PyAbstractClass
class WServiceSocketProto(WSchemeHandler):

	@abstractmethod
	@verify_type(config_selection=WConfigSelection)
	def configure_socket(self, config_selection):
		raise NotImplementedError('This method is abstract')

	@classmethod
	def scheme_specification(cls):
		""" Return scheme specification

		:return: WSchemeSpecification
		"""
		raise NotImplementedError('This method is abstract')

	@classmethod
	@verify_type(uri=WURI)
	def create_handler(cls, uri, **kwargs):
		pass


class WSocketFactoryProto(WSchemeCollection):

	@verify_type(uri=WURI, config_selection=WConfigSelection)
	def create_socket(self, uri, config_selection):
		handler = self.open(uri)
		handler.configure_socket(config_selection)
		return handler

	@verify_subclass(scheme_handler_cls=WServiceSocketProto)
	def add(self, scheme_handler_cls):
		return WSchemeCollection.add(self, scheme_handler_cls)


class WPollingHandlerProto(metaclass=ABCMeta):

	class PollingError(Exception):

		def __init__(self, *file_objects):
			Exception.__init__(self, 'Error during polling file-objects')
			self.__file_objects = file_objects

		def file_objects(self):
			return self.__file_objects

	class PollingEvent(Enum):
		read = 1
		write = 2

	def __init__(self):
		self.__file_obj = None
		self.__event_mask = None
		self.__timeout = None

	def file_obj(self):
		return tuple(self.__file_obj)

	def event_mask(self):
		return self.__event_mask

	def timeout(self):
		return self.__timeout

	def setup_poll(self, event_mask, timeout):
		if isinstance(event_mask, WPollingHandlerProto.PollingEvent) is True:
			event_mask = event_mask.value
		self.__event_mask = event_mask
		self.__timeout = timeout
		self.__file_obj = []

	def poll_fd(self, file_obj):
		if self.__file_obj is None:
			raise RuntimeError('!')
		self.__file_obj.append(file_obj)

	@abstractmethod
	def polling_function(self):
		raise NotImplementedError('This method is abstract')


class WServiceWorkerProto(metaclass=ABCMeta):

	@abstractmethod
	def process(self, socket_obj):
		raise NotImplementedError('This method is abstract')


class WServiceFactoryProto(metaclass=ABCMeta):

	@abstractmethod
	def worker(self, stop_event):
		raise NotImplementedError('This method is abstract')



# -*- coding: utf-8 -*-
#! TODO: cleanup is required

import pytest
import re

from wasp_general.network.primitives import WIPV4SocketInfo
from wasp_general.network.web.proto import WWebResponseProto, WWebPresenter, WWebTargetRouteProto, WWebRouteMapProto
from wasp_general.network.web.proto import WWebPresenterCollectionProto, WWebSessionProto, WWebServiceProto
from wasp_general.network.web.request import WWebRequest
from wasp_general.network.web.service import WSimpleErrorPresenter, WWebTargetRoute, WWebRoute, WWebRouteMap
from wasp_general.network.web.service import WWebService, WSimplePresenterCollection, WWebEnhancedPresenter
from wasp_general.network.web.service import WWebPresenterFactory
from wasp_general.network.web.headers import WHTTPHeaders
from wasp_general.network.web.debug import WWebDebugInfo


class TestWSimpleErrorPresenter:

	class Session(WWebSessionProto):

		def client_address(self):
			WIPV4SocketInfo('127.0.0.1', 65000)

		def server_address(self):
			WIPV4SocketInfo('127.0.0.1', 80)

		def protocol(self):
			return 'http'

		def protocol_version(self):
			return '0.9'

		def read_request(self):
			pass

		def write_response(self, request, response):
			pass

		def session_close(self):
			pass

	def test_error_presenter(self):
		session = TestWSimpleErrorPresenter.Session()
		request = WWebRequest(session, 'GET', '/foo')

		error = WSimpleErrorPresenter(request)
		assert(isinstance(error, WSimpleErrorPresenter) is True)
		assert(isinstance(error.error_code(501), WWebResponseProto) is True)
		assert(error.error_code(502).status() == 502)

		assert(isinstance(error.exception_error(ValueError('Text exception')), WWebResponseProto) is True)
		assert(error.exception_error(ValueError('Text exception')).status() == 500)

		name = WSimpleErrorPresenter.__presenter_name__()
		assert(isinstance(name, str) is True)
		assert(len(name) > 0)


class TestWWebTargetRoute:

	class Presenter(WWebPresenter):

		@classmethod
		def __presenter_name__(cls):
			return 'presenter name'

	class RouteMap(WWebRouteMapProto):

		def route(self, request, service):
			pass

		def error_presenter(self):
			pass

	def test_target_route(self):
		route_map = TestWWebTargetRoute.RouteMap()
		source_route = WWebRoute('/', 'presenter name')
		route = WWebTargetRoute('presenter name', 'action1', source_route, route_map, arg1='foo', arg2=3)
		assert(isinstance(route, WWebTargetRouteProto) is True)
		assert(route.route_map() == route_map)
		assert(route.presenter_name() == 'presenter name')
		assert(route.presenter_action() == 'action1')
		assert(route.presenter_args()['arg1'] == 'foo')
		assert(route.presenter_args()['arg2'] == 3)

		source_route = WWebRoute('/', 'presenter name')
		route = WWebTargetRoute('presenter name', 'action1', source_route, route_map, arg1='foo', arg2=3)
		assert(route.route() == source_route)


class TestWWebRoute:

	class Session(WWebSessionProto):

		def client_address(self):
			WIPV4SocketInfo('127.0.0.1', 65000)

		def server_address(self):
			WIPV4SocketInfo('127.0.0.1', 80)

		def protocol(self):
			return 'http'

		def protocol_version(self):
			return '0.9'

		def read_request(self):
			pass

		def write_response(self, request, response):
			pass

		def session_close(self):
			pass

	class Service(WWebServiceProto):

		def route_map(self):
			return

		def presenter_collection(self):
			return

		def presenter_factory(self):
			return

		def execute(self, request, target_route):
			return

	def test_arg_search(self):
		pytest.raises(TypeError, WWebRoute.ArgSearch)
		pytest.raises(NotImplementedError, WWebRoute.ArgSearch.args, None)
		pytest.raises(NotImplementedError, WWebRoute.ArgSearch.reduce_pattern, None)

		assert(issubclass(WWebRoute.BasicArgSearch, WWebRoute.ArgSearch) is True)
		basic_search = WWebRoute.BasicArgSearch('/section/{action}/page/{page_param1}')
		args = list(basic_search.args())
		args.sort(key=lambda x: x[1])
		assert(args == [('action', 9), ('page_param1', 23)])
		reduce_fn = basic_search.reduce_pattern()
		test_re = re.compile(reduce_fn('/foo/{bar}//1/page/'))
		assert(test_re.match('/foo/var1//1/page/') is not None)
		assert(test_re.match('/foo/1//1/page/') is not None)
		assert(test_re.match('/foo/1/page') is None)

		assert(issubclass(WWebRoute.CustomArgSearch, WWebRoute.ArgSearch) is True)
		custom_search = WWebRoute.CustomArgSearch('arg_foo', 77, '{arg_foo: "(\d+)"}', '(\d+)')
		assert(custom_search.args() == (('arg_foo', 77),))
		reduce_fn = custom_search.reduce_pattern()
		assert(reduce_fn('/zzz{arg_foo: "(\d+)"}aaa//') == '/zzz(\d+)aaa//')

		custom_search = WWebRoute.CustomArgSearch('arg1', 77, '{pattern!}', 'replace_with_text')
		reduce_fn = custom_search.reduce_pattern()
		assert(reduce_fn('/page/{pattern!}/zzz') == '/page/replace_with_text/zzz')

		pattern = '/section/{section_action: "(\w+)"}/page/{page_index: "(\d+)"}}'
		args_searches = list(WWebRoute.CustomArgSearch.custom(pattern))
		assert(args_searches[0].args() == (('section_action', 9),))
		reduce_fn = args_searches[0].reduce_pattern()
		reduce_result = reduce_fn('/section/{section_action: "(\w+)"}/page/{page_index: "(\d+)"}')
		assert(reduce_result == '/section/(\w+)/page/{page_index: "(\d+)"}')
		assert(args_searches[1].args() == (('page_index', 40),))
		reduce_fn = args_searches[1].reduce_pattern()
		reduce_result = reduce_fn('/section/{section_action: "(\w+)"}/page/{page_index: "(\d+)"}')
		assert(reduce_result == '/section/{section_action: "(\w+)"}/page/(\d+)')

	def test_route(self):
		route_map = TestWWebTargetRoute.RouteMap()
		route = WWebRoute('/foo/section/{action}/{index:"(\d+)"}/page/{page_action}', 'presenter name')
		request = WWebRequest(TestWWebRoute.Session(), 'GET', '/foo/section/bla_bla_action/13/page/show_page')

		collection = WSimplePresenterCollection()
		collection.add(TestWWebTargetRoute.Presenter)
		service = TestWWebRoute.Service()
		service.presenter_collection = lambda: collection
		service.route_map = lambda: route_map
		match = route.match(request, service)
		assert(isinstance(match, WWebTargetRoute) is True)
		assert(match.presenter_name() == 'presenter name')
		assert(match.presenter_action() == 'bla_bla_action')
		assert(match.presenter_args() == {'index': '13', 'page_action': 'show_page'})

		route = WWebRoute(
			'/index', 'presenter name', action='default_action', protocols=('http',), arg1='value1'
		)
		request = WWebRequest(TestWWebRoute.Session(), 'GET', '/index')
		match = route.match(request, service)
		assert(isinstance(match, WWebTargetRoute) is True)
		assert(match.presenter_name() == 'presenter name')
		assert(match.presenter_action() == 'default_action')
		assert(match.presenter_args()['arg1'] == 'value1')

		session = TestWWebRoute.Session()
		request = WWebRequest(session, 'GET', '/index')
		assert(route.match(request, service) is not None)

		session.protocol = lambda: 'https'
		assert(route.match(request, service) is None)

		route = WWebRoute(
			'/index//', 'presenter name', virtual_hosts=('ns.foo.bar',), methods=('POST',), ports=(8080,)
		)
		session.protocol = lambda: 'http'
		session.server_address = lambda: WIPV4SocketInfo('127.0.0.1', 8080)
		request = WWebRequest(session, 'POST', '/index//', headers=WHTTPHeaders(host='ns.foo.bar'))
		assert(route.match(request, service) is not None)

		session.server_address = lambda: WIPV4SocketInfo('127.0.0.1', 8081)
		request = WWebRequest(session, 'POST', '/index//', headers=WHTTPHeaders(host='ns.foo.bar'))
		assert(route.match(request, service) is None)

		session.server_address = lambda: WIPV4SocketInfo('127.0.0.1', 8080)
		request = WWebRequest(session, 'GET', '/index', headers=WHTTPHeaders(host='ns.foo.bar'))
		assert(route.match(request, service) is None)

		request = WWebRequest(session, 'POST', '/index')
		assert(route.match(request, service) is None)

		request = WWebRequest(session, 'POST', '/section', headers=WHTTPHeaders(host='ns.foo.bar'))
		assert(route.match(request, service) is None)

		request = WWebRequest(session, 'POST', '/index', headers=WHTTPHeaders(host='ns.foo.bar'))
		assert(route.match(request, service) is not None)

		route = WWebRoute('/index', 'presenter name', arg1='value1')
		assert(route.url_for() == '/index')

		route = WWebRoute('/index/{index}', 'presenter name', index='1')
		assert(route.url_for() == '/index/1')
		assert(route.url_for(index='page2') == '/index/page2')

		route = WWebRoute('/page/{action}/1', 'presenter name', protocols=('https',))
		assert(route.url_for(host='foo.bar') == 'https://foo.bar/page/index/1')
		assert(route.url_for(host='foo.bar', port=14) == 'https://foo.bar:14/page/index/1')
		assert(route.url_for(protocol='http', host='foo.bar') == 'http://foo.bar/page/index/1')

		route = WWebRoute('/page/{var_name}/1', 'presenter name')
		pytest.raises(RuntimeError, route.url_for)


class TestWWebRouteMap:

	class ErrorPresenter(WSimpleErrorPresenter):
		pass

	class Presenter(WWebPresenter):

		def index(self):
			pass

		@classmethod
		def __presenter_name__(cls):
			return 'presenter name'

	class SamplePresenter(WWebPresenter):

		def show(self):
			pass

		def index(self):
			pass

		@classmethod
		def __presenter_name__(cls):
			return 'SamplePresenter'

	class EnhancedPresenter(WWebEnhancedPresenter):

		def index(self):
			pass

		@classmethod
		def __presenter_name__(cls):
			return 'enhanced presenter name'

	class TargetRoute(WWebTargetRouteProto):

		def presenter_name(self):
			return 'enhanced presenter name'

		def presenter_action(self):
			return 'index'

	def test_route_map(self):
		collection = WSimplePresenterCollection()
		collection.add(TestWWebRouteMap.Presenter)
		collection.add(TestWWebRouteMap.EnhancedPresenter)
		service = TestWWebRoute.Service()
		service.presenter_collection = lambda: collection

		route_map = WWebRouteMap()
		service.route_map = lambda: route_map
		assert(isinstance(route_map, WWebRouteMap) is True)
		assert(isinstance(route_map, WWebRouteMapProto) is True)
		assert(route_map.error_presenter() == WSimpleErrorPresenter)

		route_map.set_error_presenter(TestWWebRouteMap.ErrorPresenter)
		assert(route_map.error_presenter() == TestWWebRouteMap.ErrorPresenter)

		request = WWebRequest(TestWWebRoute.Session(), 'GET', '/index')
		assert(route_map.route(request, service) is None)
		route_map.connect('/index', 'presenter name')
		assert(route_map.route(request, service) is not None)

		request = WWebRequest(TestWWebRoute.Session(), 'GET', '/page')
		assert(route_map.route(request, service) is None)
		route_map.append(WWebRoute('/page', 'presenter name'))
		assert(route_map.route(request, service) is not None)

	def test_import(self):

		collection = WSimplePresenterCollection()
		collection.add(TestWWebRouteMap.SamplePresenter)
		service = TestWWebRoute.Service()
		service.presenter_collection = lambda: collection
		route_map = WWebRouteMap()
		service.route_map = lambda: route_map

		request = WWebRequest(TestWWebRoute.Session(), 'GET', '/page/1')
		assert(route_map.route(request, service) is None)

		route_map.import_route('/page/{code:"(\d+)"} => SamplePresenter (action = show, code=404, v, foo=bar)')
		target_route = route_map.route(request, service)
		assert(target_route is not None)
		assert(target_route.presenter_args() == {'code': '1', 'v': None, 'foo': 'bar'})

		request = WWebRequest(TestWWebRoute.Session(), 'GET', '/index')
		assert(route_map.route(request, service) is None)
		route_map.import_route('/index => SamplePresenter')
		target_route = route_map.route(request, service)
		assert(target_route is not None)
		assert(target_route.presenter_args() == {})

		pytest.raises(ValueError, route_map.import_route, 'AAAA')
		pytest.raises(RuntimeError, route_map.import_route, '/page => Presenter (()')


class TestWSimplePresenterCollection:

	class DummyPresenter(WWebPresenter):

		@classmethod
		def __presenter_name__(cls):
			return 'dummy presenter'

	def test_collection(self):
		collection = WSimplePresenterCollection()
		assert(isinstance(collection, WSimplePresenterCollection) is True)
		assert(isinstance(collection, WWebPresenterCollectionProto) is True)
		assert(len(collection) == 0)

		assert(collection.presenter('Name') is None)
		assert(collection.has('Name') is False)
		assert(collection.has('dummy presenter') is False)
		collection.add(TestWWebTargetRoute.Presenter)
		collection.add(TestWSimplePresenterCollection.DummyPresenter)
		assert(collection.presenter('dummy presenter') == TestWSimplePresenterCollection.DummyPresenter)
		assert(collection.has('dummy presenter') is True)
		assert(len(collection) == 2)


class TestWWebService:

	class Collection(WSimplePresenterCollection):
		pass

	class WebRoute(WWebRouteMap):
		pass

	class Response(WWebResponseProto):

		def status(self):
			return 200

	class Presenter(WWebPresenter):

		__test__ = 1

		@classmethod
		def __presenter_name__(cls):
			return 'presenter'

		def index(self):
			return 'presenter-index'

		def add(self, a, b=5):
			return a + b

	class EnhancedPresenter(WWebEnhancedPresenter):

		def index(self, p=0):
			return TestWWebService.Response()

		@classmethod
		def __presenter_name__(cls):
			return 'enhanced presenter name'

	class Session(WWebSessionProto):

		written_responses = []

		def client_address(self):
			return

		def server_address(self):
			return

		def protocol_version(self):
			return

		def protocol(self):
			return 'http'

		def read_request(self):
			return WWebRequest(self, 'GET', '/1')

		def write_response(self, request, response):
			self.written_responses.append(response.status())

		def session_close(self):
			self.written_responses.append(-1)

	class Debugger(WWebDebugInfo):

		__requests__ = {}
		__responses__ = {}
		__target_routes__ = {}
		__exceptions__ = {}
		__finalized__ = []

		id = 0

		def session_id(self):
			self.__class__.id += 1
			self.__class__.__exceptions__[self.__class__.id] = []
			return self.__class__.id

		def request(self, session_id, request, protocol_version, protocol):
			self.__class__.__requests__[session_id] = (request, protocol_version, protocol)

		def response(self, session_id, response):
			self.__class__.__responses__[session_id] = response

		def target_route(self, session_id, target_route):
			self.__class__.__target_routes__[session_id] = target_route

		def exception(self, session_id, exc):
			self.__class__.__exceptions__[session_id].append(exc)

		def finalize(self, session_id):
			self.__class__.__finalized__.append(session_id)

	def test_service(self):

		service = WWebService()
		assert(isinstance(service.route_map(), WWebRouteMap) is True)
		assert(isinstance(service.presenter_collection(), WSimplePresenterCollection) is True)

		service = WWebService(
			route_map=TestWWebService.WebRoute(),
			collection=TestWWebService.Collection(),
			debugger=TestWWebService.Debugger()
		)
		assert(isinstance(service.route_map(), TestWWebService.WebRoute) is True)
		assert(isinstance(service.presenter_collection(), TestWWebService.Collection) is True)

		session = TestWWebService.Session()
		service.process_request(session)
		assert(TestWWebService.Session.written_responses == [404])

		service.connect('/{p}', TestWWebService.EnhancedPresenter)
		service.process_request(session)
		assert(TestWWebService.Session.written_responses == [404, 200])

		session.read_request = lambda: WWebRequest(session, 'GET', '/page/1')
		service.process_request(session)
		assert(TestWWebService.Session.written_responses == [404, 200, 404])

		class P(WWebEnhancedPresenter):

			@classmethod
			def __presenter_name__(cls):
				return 'P-presenter'

			@classmethod
			def __public_routes__(cls):
				return WWebRoute('/page/1', cls.__presenter_name__(), d=1),

			def index(self, d):
				return TestWWebService.Response()

		service.add_presenter(P)
		service.process_request(session)
		assert(TestWWebService.Session.written_responses == [404, 200, 404, 200])

		target_route = TestWWebRouteMap.TargetRoute()
		target_route.presenter_name = lambda: 'Unknown presenter'
		service.route_map().route = lambda x, y: target_route
		service.process_request(session)
		assert(TestWWebService.Session.written_responses == [404, 200, 404, 200, 500])

		service.add_presenter(TestWWebService.Presenter)

		target_route.presenter_name = lambda: 'presenter'
		target_route.presenter_action = lambda: 'unknown action'
		service.process_request(session)
		assert(TestWWebService.Session.written_responses == [404, 200, 404, 200, 500, 500])

		target_route.presenter_action = lambda: '__test__'
		service.process_request(session)
		assert(TestWWebService.Session.written_responses == [404, 200, 404, 200, 500, 500, 500])

		target_route.presenter_name = lambda: 'P-presenter'
		target_route.presenter_action = lambda: 'index'

		original_route = WWebTargetRoute(
			'P-presenter', 'index', WWebRoute('/', 'P-presenter'), service.route_map()
		)
		response = service.proxy(session.read_request(), original_route, 'presenter', action='add', a=2)
		assert(response == 7)

		def exc_fn():
			raise RuntimeError('')
		session.read_request = exc_fn
		service.process_request(session)
		assert(TestWWebService.Session.written_responses == [404, 200, 404, 200, 500, 500, 500, -1])

		requests = {}
		for key in TestWWebService.Debugger.__requests__.keys():
			requests[key] = TestWWebService.Debugger.__requests__[key][0].path()

		assert(requests == {
			1: '/1', 2: '/1', 3: '/page/1', 4: '/page/1', 5: '/page/1', 6: '/page/1', 7: '/page/1'
		})

		responses = {}
		for key in TestWWebService.Debugger.__responses__.keys():
			responses[key] = TestWWebService.Debugger.__responses__[key].status()
		assert(responses == {1: 404, 2: 200, 3: 404, 4: 200})

		targets = {}
		for key in TestWWebService.Debugger.__target_routes__.keys():
			target = TestWWebService.Debugger.__target_routes__[key]
			targets[key] = target.presenter_name() if target is not None else None
		assert(targets == {
			1: None, 2: 'enhanced presenter name', 3: None, 4: 'P-presenter',
			5: 'P-presenter', 6: 'P-presenter', 7: 'P-presenter'
		})

		exc = {}
		for key in TestWWebService.Debugger.__exceptions__.keys():
			exc[key] = [x.__class__ for x in TestWWebService.Debugger.__exceptions__[key]]
		assert(exc == {
			1: [], 2: [], 3: [], 4: [], 5: [RuntimeError], 6: [RuntimeError], 7: [RuntimeError],
			8: [RuntimeError]
		})

		assert(TestWWebService.Debugger.__finalized__ == [1, 2, 3, 4, 5, 6, 7, 8])


class TestWWebEnhancedPresenter:

	class Presenter(WWebEnhancedPresenter):

		@classmethod
		def __presenter_name__(cls):
			return 'presenter name'

	def test_presenter(self):
		assert(issubclass(WWebEnhancedPresenter, WWebPresenter) is True)

		service = WWebService()
		service.add_presenter(TestWWebService.Presenter)
		session = TestWWebService.Session()
		request = WWebRequest(session, 'GET', '/1')
		route = WWebRoute('/1', 'presenter name')
		target_route = WWebTargetRoute('presenter name', 'index', route, service.route_map())
		presenter = TestWWebEnhancedPresenter.Presenter(request, target_route, service)

		assert(presenter.__target_route__() == target_route)
		assert(presenter.__service__() == service)
		assert(WWebEnhancedPresenter.__public_routes__() == tuple())

		response = presenter.__proxy__('presenter', action='add', a=2)
		assert(response == 7)


class TestWWebPresenterFactory:

	def test_factory(self):
		factory = WWebPresenterFactory()

		class CustomPresenter(WWebPresenter):

			def __init__(self):
				return

			def __presenter_name__(cls):
				return

		assert(factory.instantiable(CustomPresenter) is False)

		service = WWebService()
		session = TestWWebService.Session()
		request = WWebRequest(session, 'GET', '/1')
		route = WWebRoute('/1', 'presenter name')
		target_route = WWebTargetRoute('presenter name', 'index', route, service.route_map())
		pytest.raises(
			RuntimeError, factory.instantiate, CustomPresenter, request, target_route, service
		)

		class CustomFactory(WWebPresenterFactory):

			def __init__(self):
				WWebPresenterFactory.__init__(self)

				def constructor(presenter_class, request, target_route, service):
					return CustomPresenter()

				self._add_constructor(CustomPresenter, constructor)

		p = CustomFactory().instantiate(CustomPresenter, request, target_route, service)
		assert(isinstance(p, CustomPresenter) is True)



# -*- coding: utf-8 -*-
# wasp_general/network/messenger/composer.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from enum import Enum

from wasp_general.verify import verify_type
from wasp_general.network.messenger.proto import WMessengerEnvelopeProto, WMessengerOnionSessionProto
from wasp_general.network.messenger.proto import WMessengerOnionLayerProto
from wasp_general.network.messenger.envelope import WMessengerEnvelope, WMessengerDictEnvelope
from wasp_general.composer import WComposerFactory


class WMessengerComposerLayer(WMessengerOnionLayerProto):

	__layer_name__ = "com.binblob.wasp-general.composer-packer-layer"
	""" Layer name
	"""

	class Mode(Enum):
		compose = 1
		decompose = 2

	def __init__(self):
		""" Construct new layer
		"""
		WMessengerOnionLayerProto.__init__(self, WMessengerComposerLayer.__layer_name__)

	@verify_type('paranoid', envelope=WMessengerEnvelopeProto, session=WMessengerOnionSessionProto, mode=Mode)
	@verify_type('paranoid', composer_factory=WComposerFactory)
	def process(self, envelope, session, mode=None, composer_factory=None,**kwargs):
		""" :meth:`.WMessengerOnionLayerProto.process` implementation
		"""
		if mode == WMessengerComposerLayer.Mode.compose:
			return self.compose(envelope, session, composer_factory, **kwargs)
		elif mode == WMessengerComposerLayer.Mode.decompose:
			return self.decompose(envelope, session, composer_factory, **kwargs)
		raise RuntimeError('Invalid mode was specified')

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerDictEnvelope, composer_factory=WComposerFactory)
	def compose(self, envelope, session, composer_factory, **kwargs):
		return WMessengerEnvelope(
			composer_factory.compose(envelope.message()),
			meta=envelope
		)

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerEnvelopeProto, composer_factory=WComposerFactory)
	def decompose(self, envelope, session, composer_factory, **kwargs):
		return WMessengerDictEnvelope(
			composer_factory.decompose(envelope.message()),
			meta=envelope
		)



# -*- coding: utf-8 -*-
# wasp_general/network/web/proto.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod, abstractclassmethod
import urllib.parse
import re

from wasp_general.verify import verify_type, verify_value, verify_subclass
from wasp_general.network.web.re_statements import http_get_vars_selection, http_post_vars_selection


class WWebResponseProto(metaclass=ABCMeta):
	""" Class represent server response for client HTTP-request
	"""

	def status(self):
		""" Return response status code. Is required for 1.0 protocol and beyond. It must be avoid
		in HTTP/0.9

		:return: None or int
		"""
		pass

	def headers(self):
		""" Return headers to set in response.

		:return: None if no headers is needed to be written or WHTTPHeaders
		"""
		pass

	def response_data(self):
		""" Return response payload

		:return: None if not required or bytes
		"""
		pass

	def __pushed_responses__(self):
		""" Return related HTTP-responses, that can be pushed to client. Available on HTTP/2 only

		:return: None if not required or tuple of WWebResponseProto
		"""
		return tuple()


class WWebRequestProto(metaclass=ABCMeta):
	""" Class represent client HTTP-request
	"""

	get_vars_re = re.compile(http_get_vars_selection)
	"""
	Regular expression that is used for GET-vars parsing
	"""

	post_vars_re = re.compile(http_post_vars_selection)
	"""
	Regular expression that is used for POST-vars parsing
	"""

	@abstractmethod
	def session(self):
		""" Return current session with which request is created

		:return: WWebSessionProto
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def method(self):
		""" Return request method (like POST/GET/...)

		:return: str
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def path(self):
		""" Return request path (like /foo/bar/index.html)

		:return: str
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def headers(self):
		""" Return request headers

		:return: None or WHTTPHeaders
		"""
		raise NotImplementedError('This method is abstract')

	def request_data(self):
		""" Return request payload

		:return: None or bytes
		"""
		pass

	def virtual_host(self):
		""" Return request virtual host ("Host" header value)

		:return: None or str
		"""
		if self.headers() is not None:
			host_value = self.headers()['Host']
			return host_value[0].lower() if host_value is not None else None

	def content_type(self):
		""" Return request content_type ("Content-Type" header value)

		:return: None or str
		"""
		if self.headers() is not None:
			return self.headers().content_type()

	def get_vars(self):
		""" Parse request path and return GET-vars

		:return: None or dictionary of names and tuples of values
		"""
		if self.method() != 'GET':
			raise RuntimeError('Unable to return get vars for non-get method')
		re_search = WWebRequestProto.get_vars_re.search(self.path())
		if re_search is not None:
			return urllib.parse.parse_qs(re_search.group(1), keep_blank_values=1)

	def post_vars(self):
		""" Parse request payload and return POST-vars

		:return: None or dictionary of names and tuples of values
		"""
		if self.method() != 'POST':
			raise RuntimeError('Unable to return post vars for non-get method')

		content_type = self.content_type()
		if content_type is None or content_type.lower() != 'application/x-www-form-urlencoded':
			raise RuntimeError('Unable to return post vars with invalid content-type request')

		request_data = self.request_data()
		request_data = request_data.decode() if request_data is not None else ''
		re_search = WWebRequestProto.post_vars_re.search(request_data)
		if re_search is not None:
			return urllib.parse.parse_qs(re_search.group(1), keep_blank_values=1)


class WWebSessionProto(metaclass=ABCMeta):
	""" Represent client session. For HTTP/0.9-1.0 every request creates new session. For HTTP/1.1 request
	can be joined in a single session if "Connection" header is used. For HTTP/2 protocol there can be single
	session for every client
	"""

	@abstractmethod
	def client_address(self):
		""" Return client IP-address and port

		:return: WIPV4SocketInfo
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def server_address(self):
		""" Return server IP-address and port to which client is connected

		:return: WIPV4SocketInfo
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def protocol_version(self):
		""" Return currently used protocol version

		:return: str (one of "0.9"/"1.0"/"1.1"/"2")
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def protocol(self):
		""" Return currently used protocol

		:return: str (one of "http"/"https")
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def read_request(self):
		""" Read next request from session

		:return: WWebRequestProto
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(request=WWebRequestProto, reponse=WWebResponseProto)
	def write_response(self, request, response):
		""" Write response to client

		:param request: original client request
		:param response: response to write
		:return:
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def session_close(self):
		""" Close session

		:return: None
		"""
		raise NotImplementedError('This method is abstract')


class WWebPresenter(metaclass=ABCMeta):
	""" Class represent worker that generates response over specified request.
	WWebTargetRouteProto defines what exactly method will be called and with what arguments.

	see :class:`.WWebTargetRouteProto`
	see :class:`.WWebRouteMapProto`
	"""

	@verify_type(request=WWebRequestProto)
	def __init__(self, request):
		""" Construct new worker

		:param request: client request
		"""
		self.__request = request

	@abstractclassmethod
	def __presenter_name__(cls):
		"""

		:return: str or None (in most cases - str)
		"""
		raise NotImplementedError('This method is abstract')

	def __request__(self):
		""" Return client request

		:return: WWebRequestProto
		"""
		return self.__request


class WWebErrorPresenter(WWebPresenter, metaclass=ABCMeta):
	""" Presenter which is used for displaying standard HTTP-errors and internal exceptions
	"""

	@abstractmethod
	@verify_type(code=int)
	@verify_value(code=lambda x: x > 0)
	def error_code(self, code):
		""" Return response for the given HTTP-code

		:param code: HTTP-code of error
		:return: WWebResponseProto
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(exception=Exception)
	def exception_error(self, exception):
		""" Return response for the given exception

		:param exception: raised exception to process and or display
		:return: WWebResponseProto
		"""
		raise NotImplementedError('This method is abstract')


class WWebTargetRouteProto(metaclass=ABCMeta):
	""" Represent single route that matches client request
	"""

	@abstractmethod
	def presenter_name(self):
		""" Return presenter name to be used

		:return: WWebPresenter
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def presenter_action(self):
		""" Return method name to be called

		:return: str
		"""
		raise NotImplementedError('This method is abstract')

	def presenter_args(self):
		""" Return arguments to be used with presenter method

		:return: dict, where keys - are argument names and values - argument values (any type)
		"""
		return dict()


class WWebPresenterCollectionProto(metaclass=ABCMeta):
	""" Represent collection of presenters

	see :class:`.WWebPresenter`
	"""

	@abstractmethod
	@verify_type(presenter_name=str)
	@verify_value(presenter_name=lambda x: len(x) > 0)
	def presenter(self, presenter_name):
		""" Return presenter by its name

		:param presenter_name: name of presenter class
		:return: type (WWebPresenter sublcass) or None (if there is no such presenter)
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(presenter_name=str)
	@verify_value(presenter_name=lambda x: len(x) > 0)
	def has(self, presenter_name):
		""" Return if there is a presenter with the given name

		:param presenter_name: name of presenter
		:return: bool
		"""
		raise NotImplementedError('This method is abstract')


class WWebPresenterFactoryProto(metaclass=ABCMeta):
	""" This class is used for presenter object instantiation. Different presenter classes may have different
	constructor and so may require different arguments, so this class implementation will be possible to
	instantiate limited presenter classes. Because of that, this class is tightly connected with specific
	WWebServiceProto implementation, which uses this class for creating presenter
	"""

	@abstractmethod
	@verify_subclass(presenter_class=WWebPresenter)
	def instantiable(self, presenter_class):
		""" Check if this factory can produce the specified presenter

		:param presenter_class: target presenter class
		:return: bool
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(presenter_name=str)
	@verify_subclass(presenter_class=WWebPresenter)
	def instantiate(self, presenter_class, *args, **kwargs):
		""" Create new presenter object. Different implementation may have different required arguments

		:param presenter_class: presenter class to instantiate
		:param args: instantiation arguments (may vary)
		:param kwargs: instantiation arguments (may vary)
		:return: WWebPresenter
		"""
		raise NotImplementedError('This method is abstract')


class WWebServiceProto(metaclass=ABCMeta):
	""" Represent service that unites wasp-general web-functionality
	"""

	@abstractmethod
	def route_map(self):
		""" Return service route map

		:return: WWebRouteMapProto
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def presenter_collection(self):
		""" Return service presenter collection

		:return: WWebPresenterCollection
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def presenter_factory(self):
		""" Return current presenter factory

		:return: WWebPresenterFactoryProto
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(request=WWebRequestProto, target_route=WWebTargetRouteProto)
	def execute(self, request, target_route):
		""" Execute given target route and return response

		:param request: client request
		:param target_route: route to execute
		:return: WWebResponseProto
		"""
		raise NotImplementedError('This method is abstract')


class WWebRouteMapProto(metaclass=ABCMeta):
	""" Represent collection of routes
	"""

	@abstractmethod
	@verify_type(request=WWebRequestProto, service=WWebServiceProto)
	def route(self, request, service):
		""" Return the first route that matches client request

		:param request: client request
		:param service: source service
		:return: None if no route is found, WWebTargetRouteProto otherwise
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def error_presenter(self):
		""" Return presenter that is used for error handling

		:return: WWebErrorPresenter
		"""
		raise NotImplementedError('This method is abstract')

	@verify_type(target_route=WWebTargetRouteProto)
	def target_route_valid(self, target_route):
		""" Check target route for execution. This method is used for omitting special methods (actions)
		from being executed

		:param target_route: route to check
		:return: bool
		"""
		return target_route.presenter_action()[0] != '_'



# -*- coding: utf-8 -*-
# wasp_general/command/command.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod
import shlex

from wasp_general.verify import verify_type
from wasp_general.command.proto import WCommandResultProto


class WCommandProto(metaclass=ABCMeta):
	""" Prototype for a single command. Command tokens are string, where each token is a part of the command name or
	is the command parameter. Tokens are generated from a string, each token is separated by space (if space is a
	part of the token, then it must be quoted). Any command may require some additional parameters that are
	generated from environment with which this command will be checked and/or called. This extra parameters
	calls command environment
	"""

	@abstractmethod
	@verify_type(command_tokens=str)
	def match(self, *command_tokens, **command_env):
		""" Checks whether this command can be called with the given tokens. Return True - if tokens match this
		command, False - otherwise

		:param command_tokens: command to check
		:param command_env: command environment
		:return: bool
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(command_tokens=str)
	def exec(self, *command_tokens, **command_env):
		""" Execute valid command (that represent as tokens)

		:param command_tokens: command to execute
		:param command_env: command environment
		:return: WCommandResultProto
		"""
		raise NotImplementedError('This method is abstract')

	@staticmethod
	@verify_type(command_str=str)
	def split_command(command_str):
		""" Split command string into command tokens

		:param command_str: command to split
		:return: tuple of str
		"""
		return shlex.split(command_str)

	@staticmethod
	@verify_type(command_tokens=str)
	def join_tokens(*command_tokens):
		""" Join tokens into a single string

		:param command_tokens: tokens to join
		:return: str
		"""
		return ' '.join([shlex.quote(x) for x in command_tokens])


class WCommand(WCommandProto):
	""" Basic WCommandProto implementation
	"""

	@verify_type(command_tokens=str)
	def __init__(self, *command_tokens):
		""" Create new command

		:param command_tokens: tokens (command) that call this command, like 'help' or ('create', 'object')
		"""
		WCommandProto.__init__(self)
		self.__command = tuple(command_tokens)

	def command(self):
		""" Return command tokens

		:return: tuple of str
		"""
		return self.__command

	@verify_type(command_tokens=str)
	def match(self, *command_tokens, **command_env):
		""" :meth:`.WCommandProto.match` implementation
		"""
		command = self.command()
		if len(command_tokens) >= len(command):
			return command_tokens[:len(command)] == command
		return False

	@abstractmethod
	@verify_type('paranoid', command_tokens=str)
	def _exec(self, *command_tokens, **command_env):
		""" Derived classes must implement this function, in order to do a real command work.

		:param command_tokens: command to execute
		:param command_env: command environment
		:return: WCommandResultProto
		"""
		raise NotImplementedError('This method is abstract')

	@verify_type(command_tokens=str)
	def exec(self, *command_tokens, **command_env):
		""" :meth:`.WCommandProto.exec` implementation

		(throws RuntimeError if tokens are invalid, and calls :meth:`.WCommand._exec` method)
		"""
		if self.match(*command_tokens, **command_env) is False:
			raise RuntimeError('Command mismatch: %s' % self.join_tokens(*command_tokens))

		return self._exec(*command_tokens, **command_env)


class WCommandSelector:
	""" This class store command and selects suitable command for the given tokens.
	"""

	def __init__(self):
		""" Create new storage/selector
		"""
		self.__commands = []

	@verify_type(command_obj=WCommandProto)
	def add(self, command_obj):
		""" Add command to selector

		:param command_obj: command to add
		:return: None
		"""
		self.__commands.append(command_obj)

	@verify_type(command_tokens=str)
	def select(self, *command_tokens, **command_env):
		""" Select suitable command, that matches the given tokens. Each new command to check is fetched with
		this object iterator (:meth:`.WCommandSelector.__iter__`)

		:param command_tokens: command
		:param command_env: command environment
		:return: WCommandProto
		"""
		for command_obj in self:
			if command_obj.match(*command_tokens, **command_env):
				return command_obj

	def __iter__(self):
		""" Iterate over internal storage and yield next command
		"""
		for command in self.__commands:
			yield command

	def __len__(self):
		""" Return command count

		:return: int
		"""
		return len(self.__commands)


class WCommandPrioritizedSelector(WCommandSelector):
	""" This class has priority for every stored commands. Command with lower priority value will be selected first.
	"""

	@verify_type(default_priority=int)
	def __init__(self, default_priority=30):
		""" Create new selector

		:param default_priority: priority for commands, that were added via \
		:meth:`.WCommandPrioritizedSelector.add` method
		"""
		WCommandSelector.__init__(self)
		self.__default_priority = default_priority
		self.__priorities = {}

	@verify_type(command_obj=WCommandProto)
	def add(self, command_obj):
		""" :meth:`.WCommandSelector.add` redefinition (sets default priority for the given command)
		"""
		self.add_prioritized(command_obj, self.__default_priority)

	@verify_type(command_obj=WCommandProto, priority=int)
	def add_prioritized(self, command_obj, priority):
		""" Add command with the specified priority

		:param command_obj: command to add
		:param priority: command priority
		:return: None
		"""
		if priority not in self.__priorities.keys():
			self.__priorities[priority] = []

		self.__priorities[priority].append(command_obj)

	def __iter__(self):
		""" Iterate over internal storage and yield next command. Commands with lower priority will be yielded
		first
		"""
		priorities = list(self.__priorities.keys())
		priorities.sort()

		for priority in priorities:
			for command in self.__priorities[priority]:
				yield command

	def __len__(self):
		""" Return command count

		:return: int
		"""
		result = 0
		for commands in self.__priorities.values():
			result += len(commands)
		return result


class WCommandSet:
	""" Class wraps routine of execution command from a command group. This class is able to keep command
	environment variables from previous commands results to use them in a future commands calls. Only
	those variables whose names were specified in a constructor will be kept.
	"""

	class NoCommandFound(Exception):
		""" Exception that is raised when no suitable command was found during :meth:`.WCommandSet.exec` method
		"""
		pass

	@verify_type(command_selector=(WCommandSelector, None), follow_vars=(list, tuple, set, None))
	def __init__(self, command_selector=None, tracked_vars=None):
		""" Create new set

		:param command_selector: selector (storage) for commands to use
		:param tracked_vars: if it is specified - tuple/list/set of variables names, that must be kept \
		between commands calls
		"""
		self.__commands = command_selector if command_selector is not None else WCommandSelector()
		self.__tracked_vars = tuple(tracked_vars) if tracked_vars is not None else tuple()
		self.__vars = {}

	def commands(self):
		""" Return used command selector

		:return: WCommandSelector
		"""
		return self.__commands

	def tracked_vars(self):
		""" Return variables names that are kept (tracked) by this command set

		:return: tuple of str
		"""
		return self.__tracked_vars

	def has_var(self, var_name):
		""" Return True - if a environment variable with a specified name is kept by this command set.
		Otherwise - False is returned

		:param var_name: variable name to check

		:return: bool
		"""
		return var_name in self.__vars.keys()

	def var_value(self, var_name):
		""" Return value of environment variable that is kept by this command set.

		:note: No checks are made if there is a such variable. It implies that there is a such variable.
		For any doubt - use :meth:`.WCommandSet.has_var` method

		:param var_name: target variable name

		:return: anything
		"""
		return self.__vars[var_name]

	@verify_type('paranoid', command_str=str)
	def exec(self, command_str, **command_env):
		""" Execute the given command (command will be split into tokens, every space that is a part of a token
		must be quoted)

		:param command_str: command to execute
		:param command_env: command environment
		:return: WCommandResultProto
		"""
		env = self.__vars.copy()
		env.update(command_env)

		command_tokens = WCommandProto.split_command(command_str)
		command_obj = self.commands().select(*command_tokens, **env)
		if command_obj is None:
			raise WCommandSet.NoCommandFound('No suitable command found: "%s"' % command_str)

		result = command_obj.exec(*command_tokens, **env)
		self.__track_vars(result)
		return result

	@verify_type(command_result=WCommandResultProto)
	def __track_vars(self, command_result):
		""" Check if there are any tracked variable inside the result. And keep them for future use.

		:param command_result: command result tot check

		:return:
		"""
		command_env = command_result.environment()
		for var_name in self.tracked_vars():
			if var_name in command_env.keys():
				self.__vars[var_name] = command_env[var_name]


class WCommandAlias(WCommandProto):
	""" Prototype for command, that doesn't do anything useful itself, but it helps to run a modified command
	and to return its result. For getting any useful result a :class:`.WCommandSelector` object is used.
	"""

	@verify_type(selector=WCommandSelector)
	def __init__(self, selector):
		""" Create new command alias

		:param selector: selector that has commands, that will be run from this one
		"""
		WCommandProto.__init__(self)
		self.__selector = selector

	def selector(self):
		""" Return original command selector

		:return: WCommandSelector
		"""
		return self.__selector

	@abstractmethod
	def mutate_command_tokens(self, *command_tokens):
		""" Modify the input command so it can be called from the command selector

		:param command_tokens:
		:return:
		"""
		raise NotImplementedError('This method is abstract')

	@verify_type(command_tokens=str)
	def match(self, *command_tokens, **command_env):
		""" :meth:`.WCommandProto.match` implementation
		"""
		mutated_command_tokens = self.mutate_command_tokens(*command_tokens)
		if mutated_command_tokens is None:
			return False
		return self.selector().select(*mutated_command_tokens, **command_env) is not None

	@verify_type(command_tokens=str)
	def exec(self, *command_tokens, **command_env):
		""" :meth:`.WCommandProto.exec` implementation
		"""
		mutated_command_tokens = self.mutate_command_tokens(*command_tokens)
		if mutated_command_tokens is not None:
			command = self.selector().select(*mutated_command_tokens, **command_env)
			if command is not None:
				return command.exec(*mutated_command_tokens, **command_env)

		raise RuntimeError('Command mismatch: %s' % self.join_tokens(*command_tokens))


class WReduceCommand(WCommandAlias):
	""" Command that creates subsection from a command selector. The command will be matched to command tokens only
	if the first token matches to one of reduce tokens (section name/aliases) and the command selector has command
	for remaining tokens. Command execution works the same way. The command will be executed from the command
	selector without the first token and only if the first token matches to one of reduce tokens.
	"""

	@verify_type('paranoid', selector=WCommandSelector)
	@verify_type(reduce_tokens=str)
	def __init__(self, selector, *reduce_tokens):
		""" Create new command

		:param selector: selector to use
		:param reduce_tokens: section names (aliases)
		"""
		WCommandAlias.__init__(self, selector)
		if len(reduce_tokens) == 0:
			raise RuntimeError('No reduce tokens are specified')
		self.__reduce_tokens = reduce_tokens

	def reduce_tokens(self):
		""" Return section names (aliases)

		:return: tuple of str
		"""
		return self.__reduce_tokens

	def mutate_command_tokens(self, *command_tokens):
		""" :meth:`.WCommandAlias.mutate_command_tokens` implementation
		"""
		if len(command_tokens) > 0:
			if command_tokens[0] in self.reduce_tokens():
				return command_tokens[1:]



# -*- coding: utf-8 -*-
# wasp_general/network/web/debug.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type
from wasp_general.network.web.proto import WWebRequestProto, WWebResponseProto, WWebTargetRouteProto


class WWebDebugInfo(metaclass=ABCMeta):
	""" This is API prototype for web-service debugging process
	"""

	@abstractmethod
	def session_id(self):
		""" Create new token, that is used for session identification

		:return: any type
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(request=WWebRequestProto, protocol_version=str, protocol=str)
	def request(self, session_id, request, protocol_version, protocol):
		""" Dump client request

		:param session_id: session origin
		:param request: client request
		:param protocol_version: client protocol version (like 0.9/1.0/1.1/2)
		:param protocol: client protocol (like http/https)
		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(response=WWebResponseProto)
	def response(self, session_id, response):
		""" Dump server response to client

		:param session_id: session origin
		:param response: server response
		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(target_route=(WWebTargetRouteProto, None))
	def target_route(self, session_id, target_route):
		""" Dump target route

		:param session_id: session origin
		:param target_route: target route
		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	@verify_type(exc=Exception)
	def exception(self, session_id, exc):
		""" Dump raised exception (may be called more then once)

		:param session_id: session origin
		:param exc: raised exception
		:return: None
		"""
		raise NotImplementedError('This method is abstract')

	@abstractmethod
	def finalize(self, session_id):
		""" Client session finalization. This method is called whenever exceptions were risen or not

		:param session_id: session origin
		:return: None
		"""
		raise NotImplementedError('This method is abstract')



# -*- coding: utf-8 -*-
# wasp_general/network/messenger/session.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import re
from abc import ABCMeta, abstractmethod

from wasp_general.verify import verify_type
from wasp_general.network.messenger.proto import WMessengerEnvelopeProto, WMessengerOnionSessionFlowProto
from wasp_general.network.messenger.proto import WMessengerOnionSessionProto, WMessengerOnionProto
from wasp_general.network.messenger.envelope import WMessengerTextEnvelope, WMessengerBytesEnvelope


class WMessengerOnionSessionFlow(WMessengerOnionSessionFlowProto):
	""" Simple :class:`.WMessengerOnionSessionFlowProto` implementation. This class returns iterator, that was
	given in constructor
	"""

	@verify_type(iterator=WMessengerOnionSessionFlowProto.Iterator)
	def __init__(self, iterator):
		""" Construct new session flow

		:param iterator: flow iterator
		"""
		WMessengerOnionSessionFlowProto.__init__(self)
		self.__iterator = iterator

	@verify_type('paranoid', envelope=WMessengerEnvelopeProto)
	def iterator(self, envelope):
		""" :meth:`.WMessengerOnionSessionFlowProto.iterator` implementation
		"""
		return self.__iterator

	@classmethod
	@verify_type(info=WMessengerOnionSessionFlowProto.IteratorInfo)
	def sequence(cls, *info):
		""" Useful method to generate iterator. It is generated by chaining the given info. If no info is
		specified, then None is returned

		:param info: iterator info sequence
		:return: WMessengerOnionSessionFlowProto.Iterator or None
		"""
		if len(info) == 0:
			return

		info = list(info)
		info.reverse()

		result = WMessengerOnionSessionFlowProto.Iterator(
			info[0].layer_name(), **info[0].layer_args()
		)

		for i in range(1, len(info)):
			result = WMessengerOnionSessionFlowProto.Iterator(
				info[i].layer_name(), next_iterator=result, **info[i].layer_args()
			)

		return result

	@staticmethod
	@verify_type('paranoid', info=WMessengerOnionSessionFlowProto.IteratorInfo)
	def sequence_flow(*info):
		""" Same method as :meth:`.WMessengerOnionSessionFlow.sequence`, but as a result
		:class:`.WMessengerOnionSessionFlow` object created (if no info is specified, then None is returned)

		:param info: iterator info sequence
		:return: WMessengerOnionSessionFlow or None
		"""
		if len(info) > 0:
			return WMessengerOnionSessionFlow(WMessengerOnionSessionFlow.sequence(*info))


class WMessengerOnionSessionFlowSequence(WMessengerOnionSessionFlowProto):
	""" Session flow that "iterates" over flows, that was given in constructor. So this class returns iterator, that
	smoothly iterates a iterator from the first given flow, then from the second one, and so on. If no flows are
	specified in a constructor, then :meth:`.WMessengerOnionSessionFlowSequence.iterator` method returns None.
	In :meth:`.WMessengerOnionSessionFlowSequence.iterator` method no iterator is returned if all of the given
	flows doesn't return iterator for the specified envelope.
	"""

	class FlowSequenceIterator(WMessengerOnionSessionFlowProto.Iterator):
		""" Iterator that has it's own information and sequence of flows to iterate. During iteration (in
		:meth:`.WMessengerOnionSessionFlowSequence.FlowSequenceIterator.next` method) this class doesn't return
		the following iterator, but returns itself instead. By each
		:meth:`.WMessengerOnionSessionFlowSequence.FlowSequenceIterator.next` call this object saves a "real"
		iterator and saves its state (layer name and arguments), so when layer information is requested,
		information from the "real" iterator is returned
		"""
		@verify_type(info=WMessengerOnionSessionFlowProto.IteratorInfo, flows=WMessengerOnionSessionFlowProto)
		def __init__(self, info, *flows):
			""" Construct new iterator

			:param info: self information
			:param flows: flows to iterate sequentially
			"""
			WMessengerOnionSessionFlowProto.Iterator.__init__(self, info.layer_name(), **info.layer_args())

			self.__flows = list(flows)
			self.__current_flow = None
			self.__iterator = None

		@verify_type('paranoid', envelope=WMessengerEnvelopeProto)
		def __next_flow(self, envelope):
			""" Checks current state and finds next suitable "real" iterator

			:param envelope: original envelope from \
			:meth:`.WMessengerOnionSessionFlowSequence.FlowSequenceIterator.next` method
			:return: None
			"""
			self.__current_flow = (self.__current_flow + 1) if self.__current_flow is not None else 0
			if self.__current_flow < len(self.__flows):
				self.__iterator = self.__flows[self.__current_flow].iterator(envelope)
				if self.__iterator is None:
					self.__next_flow(envelope)

		@verify_type('paranoid', envelope=WMessengerEnvelopeProto)
		def next(self, envelope):
			""" :meth:`.WMessengerOnionSessionFlowProto.Iterator.next` implementation
			"""
			if self.__iterator is not None:
				self.__iterator = self.__iterator.next(envelope)

			if self.__iterator is None:
				self.__next_flow(envelope)

			if self.__iterator is not None:
				return self

		def layer_name(self):
			""" :meth:`.WMessengerOnionSessionFlowProto.IteratorInfo.layer_name` implementation
			"""
			if self.__current_flow is None:
				return WMessengerOnionSessionFlowProto.Iterator.layer_name(self)
			return self.__iterator.layer_name()

		def layer_args(self):
			""" :meth:`.WMessengerOnionSessionFlowProto.IteratorInfo.layer_args` implementation
			"""
			if self.__current_flow is None:
				return WMessengerOnionSessionFlowProto.Iterator.layer_args(self)
			return self.__iterator.layer_args()

	@verify_type(flows=WMessengerOnionSessionFlowProto)
	def __init__(self, *flows):
		""" Construct new session flow

		:param flows: flows to iterate
		"""
		WMessengerOnionSessionFlowProto.__init__(self)
		self.__flows = flows

	@verify_type('paranoid', envelope=WMessengerEnvelopeProto)
	def iterator(self, envelope):
		""" :meth:`WMessengerOnionSessionFlowProto.iterator` implementation
		"""
		iterator = WMessengerOnionSessionFlowSequence.FlowSequenceIterator(
			WMessengerOnionSessionFlowProto.IteratorInfo(''), *self.__flows
		)
		return iterator.next(envelope)


class WMessengerOnionSessionFlexFlow(WMessengerOnionSessionFlowProto):
	""" Flexible flow that choose one of the given flows depends on the given envelope. Flows are passed as pairs
	(pair of flow and object that checks the envelope) and as default flow (flow that is selected, if envelope
	wasn't "matched" to any flow)
	"""

	class MessageComparator(metaclass=ABCMeta):
		""" Abstract class that checks envelope
		"""

		@abstractmethod
		@verify_type(envelope=WMessengerEnvelopeProto)
		def match(self, envelope):
			""" Check if envelope is suitable

			:param envelope: envelope to check
			:return: bool
			"""
			raise NotImplementedError('This method is abstract')

	class ReComparator(MessageComparator):
		""" Simple :class:`.WMessengerOnionSessionFlexFlow.MessageComparator` implementation, that
		checks if envelope (text or binary) matches the given regular expression
		"""

		@verify_type(pattern=(str, bytes))
		def __init__(self, pattern):
			""" Construct new comparator

			:param pattern: regular expression to check
			"""
			WMessengerOnionSessionFlexFlow.MessageComparator.__init__(self)
			self.__re = re.compile(pattern)

		@verify_type(envelope=(WMessengerTextEnvelope, WMessengerBytesEnvelope))
		def match(self, envelope):
			""" :meth:`.WMessengerOnionSessionFlexFlow.MessageComparator.match` implementation
			"""
			return self.__re.match(envelope.message()) is not None

	class FlowComparatorPair:
		""" Pair that links comparator and flow
		"""

		@verify_type(flow=WMessengerOnionSessionFlowProto)
		def __init__(self, comparator, flow):
			""" Construct new pair

			:param comparator: comparator, that checks envelope
			:param flow: flow to use
			"""
			if isinstance(comparator, WMessengerOnionSessionFlexFlow.MessageComparator) is False:
				raise TypeError('Invalid type for comparator argument')

			self.__comparator = comparator
			self.__flow = flow

		def comparator(self):
			""" Return comparator

			:return: WMessengerOnionSessionFlexFlow.MessageComparator
			"""
			return self.__comparator

		def flow(self):
			""" Return flow

			:return: WMessengerOnionSessionFlowProto
			"""
			return self.__flow

	@verify_type(default_flow=(WMessengerOnionSessionFlowProto, None))
	def __init__(self, *flow_comparator_pairs, default_flow=None):
		""" Construct new session flow

		:param flow_comparator_pairs: flows to check
		:param default_flow: default flow (if no flow is matched)
		"""
		WMessengerOnionSessionFlowProto.__init__(self)
		self.__pairs = []
		for pair in flow_comparator_pairs:
			if isinstance(pair, WMessengerOnionSessionFlexFlow.FlowComparatorPair) is False:
				raise TypeError('Invalid type for flow-comparator pair argument')
			self.__pairs.append(pair)
		self.__default_flow = default_flow

	@verify_type('paranoid', envelope=WMessengerEnvelopeProto)
	def iterator(self, envelope):
		""" :meth:`WMessengerOnionSessionFlowProto.iterator` implementation
		"""
		for pair in self.__pairs:
			if pair.comparator().match(envelope) is True:
				return pair.flow().iterator(envelope)

		if self.__default_flow is not None:
			return self.__default_flow.iterator(envelope)


class WMessengerOnionSession(WMessengerOnionSessionProto):
	""" :class:`.WMessengerOnionSessionProto` class implementation. This class executes layers from onion in order
	they described in :class:`.WMessengerOnionSessionFlow` class
	"""

	@verify_type(onion=WMessengerOnionProto, session_flow=WMessengerOnionSessionFlowProto)
	def __init__(self, onion, session_flow):
		""" Construct new session.

		:param onion: related onion
		:param session_flow: defines layers order to be executed
		"""
		self.__onion = onion
		self.__session_flow = session_flow

	def onion(self):
		""" :meth:`.WMessengerOnionSessionProto.onion` method implementation.
		"""
		return self.__onion

	def session_flow(self):
		""" Return related :class:`.WMessengerOnionSessionFlow` object
		:return: WMessengerOnionSessionFlow
		"""
		return self.__session_flow

	@verify_type('paranoid', envelope=WMessengerEnvelopeProto)
	def process(self, envelope):
		""" :meth:`.WMessengerOnionSessionProto.process` method implementation.
		"""

		def process_single_layer(iter_envelope, iter_obj):
			layer = self.onion().layer(iter_obj.layer_name())
			layer_args = iter_obj.layer_args()
			return layer.process(iter_envelope, self, **layer_args)

		iterator = self.session_flow().iterator(envelope)
		if iterator is not None:
			while iterator is not None:
				envelope = process_single_layer(envelope, iterator)
				iterator = iterator.next(envelope)
		return envelope



# -*- coding: utf-8 -*-
# wasp_general/network/primitives.py
#
# Copyright (C) 2016 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import re

from wasp_general.verify import verify_type, verify_value

from wasp_general.types.binarray import WBinArray
from wasp_general.types.bytearray import WFixedSizeByteArray


class WMACAddress:
	""" Represent Ethernet/WiFi MAC address.

	see also https://en.wikipedia.org/wiki/MAC_address
	"""

	octet_count = 6
	""" Address bytes count
	"""

	re_dash_format = re.compile(
		"^[0-9a-fA-F]{2}-[0-9a-fA-F]{2}-[0-9a-fA-F]{2}-[0-9a-fA-F]{2}-[0-9a-fA-F]{2}-[0-9a-fA-F]{2}$"
	)
	""" Regular expression for "dash"-written MAC address like '00-11-22-33-44-55'
	"""
	re_colon_format = re.compile(
		"^[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}:[0-9a-fA-F]{2}$"
	)
	""" Regular expression for "colon"-written MAC address like '00:11:22:33:44:55'
	"""
	re_cisco_format = re.compile("^[0-9a-fA-F]{4}.[0-9a-fA-F]{4}.[0-9a-fA-F]{4}$")
	""" Regular expression for MAC address written in "Cisco" style like '0011.2233.4455'
	"""
	re_spaceless_format = re.compile("^[0-9a-fA-F]{12}$")
	""" Regular expression for MAC address written without separator like '001122334455'
	"""

	@verify_type('paranoid', address=(str, WBinArray, int, None))
	def __init__(self, address=None):
		""" Construct new address

		:param address: value with which this address is initialized (zeroes are used by default).
		"""
		self.__address = WFixedSizeByteArray(WMACAddress.octet_count)

		if address is not None:
			if isinstance(address, str):
				self.__address = WFixedSizeByteArray(
					WMACAddress.octet_count, WMACAddress.from_string(address).bin_address()
				)
			else:
				self.__address = WFixedSizeByteArray(WMACAddress.octet_count, address)

	def bin_address(self):
		""" Get this address as sequence of bits

		:return: WBinArray
		"""
		return self.__address.bin_value()

	@staticmethod
	@verify_type(address=str)
	def from_string(address):
		""" Return new object by the given MAC-address

		:param address: address to convert
		:return: WMACAddress
		"""
		str_address = None

		if WMACAddress.re_dash_format.match(address):
			str_address = "".join(address.split("-"))
		elif WMACAddress.re_colon_format.match(address):
			str_address = "".join(address.split(":"))
		elif WMACAddress.re_cisco_format.match(address):
			str_address = "".join(address.split("."))
		elif WMACAddress.re_spaceless_format.match(address):
			str_address = address

		if str_address is None:
			raise ValueError("Invalid MAC address format: " + address)

		result = WMACAddress()
		for octet_index in range(WMACAddress.octet_count):
			octet = str_address[:2]
			result.__address[octet_index] = int(octet, 16)
			str_address = str_address[2:]

		return result

	def __str__(self):
		""" Convert to string

		:return: str
		"""
		address = ["{:02x}".format(int(x)) for x in self.__address]
		return ':'.join(address)

	def __bytes__(self):
		""" Convert to bytes

		:return: bytes
		"""
		return bytes(self.__address)


class WIPV4Address:
	""" Represent IPv4 address.

	see also https://en.wikipedia.org/wiki/IPv4#Address_representations
	"""

	octet_count = 4
	""" Address bytes count
	"""

	@verify_type('paranoid', address=(str, bytes, WBinArray, int, None))
	def __init__(self, address=None):
		""" Create new address

		:param address: value with which this address is initialized (zeroes are used by default).
		"""
		self.__address = WFixedSizeByteArray(WIPV4Address.octet_count)

		if address is not None:
			if isinstance(address, str):
				self.__address = WFixedSizeByteArray(
					WIPV4Address.octet_count, WIPV4Address.from_string(address).bin_address()
				)
			else:
				self.__address = WFixedSizeByteArray(WIPV4Address.octet_count, address)

	def bin_address(self):
		""" Convert address to WBinArray

		:return: WBinArray
		"""
		return self.__address.bin_value()

	def __bytes__(self):
		""" Convert address to bytes

		:return: bytes
		"""
		return bytes(self.__address)

	@staticmethod
	@verify_type(address=str)
	def from_string(address):
		""" Parse string for IPv4 address

		:param address: address to parse
		:return:
		"""
		address = address.split('.')
		if len(address) != WIPV4Address.octet_count:
			raise ValueError('Invalid ip address: %s' % address)

		result = WIPV4Address()
		for i in range(WIPV4Address.octet_count):
			result.__address[i] = WBinArray(int(address[i]), WFixedSizeByteArray.byte_size)
		return result

	@staticmethod
	@verify_type(dns_format=bool)
	def to_string(address, dns_format=False):
		""" Convert address to string

		:param address: WIPV4Address to convert
		:param dns_format: whether to use arpa-format or not
		:return:
		"""
		if isinstance(address, WIPV4Address) is False:
			raise TypeError('Invalid address type')

		address = [str(int(x)) for x in address.__address]
		if dns_format is False:
			return '.'.join(address)
		address.reverse()
		return ('.'.join(address) + '.in-addr.arpa')

	def __str__(self):
		return WIPV4Address.to_string(self)


class WNetworkIPV4:
	""" This class represent IPv4 network address. Depends on a flag network_address (that is passed to constructor)
	object can represent separate host address with network mask or be an address of a IP network.

	see also https://en.wikipedia.org/wiki/IPv4
	"""

	delimiter = '/'
	""" Separator that separate network address from network mask.
	"""

	@verify_type(address=(str, tuple, list, set), network_address=bool)
	@verify_value(address=lambda x: isinstance(x, (tuple, list, set)) is False or len(x) == 2)
	@verify_value(address=lambda x: isinstance(x, (tuple, list, set)) is False or isinstance(x[0], WIPV4Address))
	@verify_value(address=lambda x: isinstance(x, (tuple, list, set)) is False or isinstance(x[1], int))
	def __init__(self, address, network_address=True):
		""" Construct new network address

		:param address: address as a string or as tuple/list/set. where the first element is WIPV4Address and \
		the second one is network mask (as bits count)
		:param network_address: whether this object is host address with network mask or an address of a network
		"""

		if isinstance(address, str) is True:
			ip_address, self.__mask = address.strip().split(WNetworkIPV4.delimiter)
			self.__address = WIPV4Address(ip_address)
		else:
			self.__address = address[0]
			self.__mask = address[1]

		self.__mask = int(self.__mask)
		if self.__mask < 0 or self.__mask > len(self.__address.bin_address()):
			raise ValueError(
				'Invalid network mask: %s (network specified: %s)' % (str(self.__mask), address)
			)

		self.__network_address__ = network_address
		if network_address:
			bin_address = self.__address.bin_address()
			if int(bin_address[self.__mask:]) != 0:
				raise ValueError(
					'Invalid network mask: %s (network specified: %s)' % (str(self.__mask), address)
				)

	def address(self):
		""" Return IP address

		:return: WIPV4Address
		"""
		return self.__address

	def mask(self):
		""" Return network mask (as bits count)

		:return: int
		"""
		return self.__mask

	@verify_type(skip_network_address=bool)
	def first_address(self, skip_network_address=True):
		""" Return the first IP address of this network

		:param skip_network_address: this flag specifies whether this function returns address of the network \
		or returns address that follows address of the network (address, that a host could have)
		:return: WIPV4Address
		"""
		bin_address = self.__address.bin_address()
		bin_address_length = len(bin_address)
		if self.__mask > (bin_address_length - 2):
			skip_network_address = False

		for i in range(bin_address_length - self.__mask):
			bin_address[self.__mask + i] = 0

		if skip_network_address:
			bin_address[bin_address_length - 1] = 1

		return WIPV4Address(bin_address)

	@verify_type(skip_broadcast_address=bool)
	def last_address(self, skip_broadcast_address=True):
		""" Return the last IP address of this network

		:param skip_broadcast_address: this flag specifies whether to skip the very last address (that is \
		usually used as broadcast address) or not.
		:return: WIPV4Address
		"""
		bin_address = self.__address.bin_address()
		bin_address_length = len(bin_address)
		if self.__mask > (bin_address_length - 2):
			skip_broadcast_address = False

		for i in range(bin_address_length - self.__mask):
			bin_address[self.__mask + i] = 1

		if skip_broadcast_address:
			bin_address[bin_address_length - 1] = 0

		return WIPV4Address(bin_address)

	def iterator(self, skip_network_address=True, skip_broadcast_address=True):
		""" Return iterator, that can iterate over network addresses

		:param skip_network_address: same as skip_network_address in :meth:`.NetworkIPV4.first_address` method
		:param skip_broadcast_address: same as skip_broadcast_address in :meth:`.NetworkIPV4.last_address` \
		method
		:return: NetworkIPV4Iterator
		"""
		return WNetworkIPV4Iterator(self, skip_network_address, skip_broadcast_address)

	@verify_type(address=WIPV4Address)
	def __contains__(self, address):
		""" Check if this network contains specified IP address.

		:param address: address to check
		:return: bool
		"""
		int_value = int(address.bin_address())
		first_address = int(self.first_address().bin_address())
		last_address = int(self.last_address().bin_address())
		return int_value >= first_address and int_value <= last_address

	@staticmethod
	@verify_type(address=WIPV4Address)
	def is_multicast(address):
		""" Check if address is a multicast address.

		:param address: IP address to check
		:return: bool

		see also https://tools.ietf.org/html/rfc5771
		"""
		return address in WNetworkIPV4('224.0.0.0/4')


class WNetworkIPV4Iterator:
	""" This iterator iterates over IP network addresses.
	"""

	@verify_type(network=(WIPV4Address, WNetworkIPV4), skip_network_address=bool, skip_broadcast_address=bool)
	def __init__(self, network, skip_network_address=True, skip_broadcast_address=True):
		""" Create new iterator

		:param network: network to iterate. If it is WIPV4Address instance, then iterate over single address
		:param skip_network_address: same as skip_network_address in :meth:`.NetworkIPV4.first_address` method
		:param skip_broadcast_address: same as skip_broadcast_address in :meth:`.NetworkIPV4.last_address` \
		method
		"""
		if isinstance(network, WIPV4Address) is True:
			self.__network = WNetworkIPV4((network, len(network.bin_address())))
		else:
			# isinstance(network, NetworkIPV4) is True
			self.__network = network

		self.__skip_network_address = skip_network_address
		self.__skip_broadcast_address = skip_broadcast_address

	def __iter__(self):
		""" Iterate call

		:return: None
		"""
		first_address = self.__network.first_address(skip_network_address=self.__skip_network_address)
		last_address = self.__network.last_address(skip_broadcast_address=self.__skip_broadcast_address)

		for i in range(int(first_address.bin_address()), int(last_address.bin_address()) + 1):
			yield WIPV4Address(i)


class WIPPort:
	""" Represent TCP/UDP IP port

	see also:
	https://en.wikipedia.org/wiki/Transmission_Control_Protocol#TCP_ports
	https://en.wikipedia.org/wiki/User_Datagram_Protocol#Service_ports
	"""

	minimum_port_number = 1
	""" Minimum port number
	"""
	maximum_port_number = 65535
	""" Maximum port number
	"""

	@verify_type(port=int)
	@verify_value(port=lambda x: WIPPort.minimum_port_number <= x <= WIPPort.maximum_port_number)
	def __init__(self, port):
		""" Construct new IP port

		:param port: initialization value
		"""
		self.__port = port

	def __int__(self):
		""" Return port number as int
		:return: int
		"""
		return self.__port

	def __str__(self):
		""" Return port number as string
		:return: str
		"""
		return str(self.__port)


class WFQDN:
	""" Represent single fully qualified domain name (FQDN).

	see also https://en.wikipedia.org/wiki/Fully_qualified_domain_name
	"""

	re_label = re.compile('^[a-zA-Z0-9\-]{1,63}$')
	""" Regular expression for FQDN label (sequence between dots) as is specified in ....
	"""

	maximum_fqdn_length = 253

	@verify_type(address=(str, list, tuple, set, None))
	def __init__(self, address=None):
		""" Construct new FQDN. If no address is specified, then this FQDN represent root node, which is '.'

		:param address: FQDN address in string or in list/tuple/set of labels
		"""
		self._labels = []

		if isinstance(address, str) is True:
			self._labels = WFQDN.from_string(address)._labels
		elif isinstance(address, (list, tuple, set)) is True:
			self._labels = WFQDN.from_string('.'.join(address))._labels

	@staticmethod
	@verify_type(address=str)
	def from_string(address):
		""" Convert doted-written FQDN address to WFQDN object

		:param address: address to convert
		:return: WFQDN
		"""
		if len(address) == 0:
			return WFQDN()

		if address[-1] == '.':
			address = address[:-1]

		if len(address) > WFQDN.maximum_fqdn_length:
			raise ValueError('Invalid address')

		result = WFQDN()
		for label in address.split('.'):
			if isinstance(label, str) and WFQDN.re_label.match(label):
				result._labels.append(label)
			else:
				raise ValueError('Invalid address')

		return result

	@staticmethod
	@verify_type(leading_dot=bool)
	def to_string(address, leading_dot=False):
		""" Return doted-written address by the given WFQDN object

		:param address: address to convert
		:param leading_dot: whether this function place leading dot to the result or not
		:return: str
		"""

		if isinstance(address, WFQDN) is False:
			raise TypeError('Invalid type for FQDN address')

		result = '.'.join(address._labels)
		return result if leading_dot is False else (result + '.')

	def __str__(self):
		""" Return string

		:return: str
		"""
		return WFQDN.to_string(self)

	@staticmethod
	@verify_type(idn_fqdn=str)
	def punycode(idn_fqdn):
		""" Create WFQDN from IDN (Internationalized domain name) by reverting it to punycode

		:param idn_fqdn: internationalized domain name to convert
		:return: WFQDN

		see also https://en.wikipedia.org/wiki/Internationalized_domain_name
		see also https://en.wikipedia.org/wiki/Punycode
		"""
		return WFQDN(idn_fqdn.encode('idna').decode('ascii'))


class WIPV4SocketInfo:
	""" Represent socket information - IP address (or domain name) and port number. Mainly used for python socket
	module.

	see :meth:`.WIPV4SocketInfo.pair`
	"""

	@verify_type('paranoid', address=(WFQDN, WIPV4Address, str, WBinArray, int, None), port=(WIPPort, int, None))
	def __init__(self, address=None, port=None):
		""" Construct new pair

		:param address: associated IP address
		:param port: associated IP port
		"""
		self.__address = None
		if address is not None:
			if isinstance(address, (WFQDN, WIPV4Address)) is True:
				self.__address = address
			elif isinstance(address, (WBinArray, int)) is True:
				self.__address = WIPV4Address(address)
			else:
				self.__address = WIPV4SocketInfo.parse_address(address)

		self.__port = None
		if port is not None:
			self.__port = port if isinstance(port, WIPPort) else WIPPort(port)

	def address(self):
		""" Return associated IP address or None if not available

		:return: WIPV4Address or WFQDN or None
		"""
		return self.__address

	def port(self):
		""" Return associated IP port or None if not available

		:return: WIPPort or None
		"""
		return self.__port

	def pair(self):
		""" Return tuple (address, port), where address is a string (empty string if self.address() is None) and
		port is an integer (zero if self.port() is None). Mainly, this tuple is used with python socket module
		(like in bind method)

		:return: 2 value tuple of str and int.
		"""
		address = str(self.__address) if self.__address is not None else ''
		port = int(self.__port) if self.__port is not None else 0
		return address, port

	@staticmethod
	@verify_type(address=str)
	def parse_address(address):
		""" Parse string and return :class:`.WIPV4Address` object if an IP address is specified,
		:class:`.WFQDN` if domain name is specified and None if the string is empty.

		:param address: string to parse
		:return: WIPV4Address or WFQDN or None
		"""

		if len(address) == 0:
			return None

		try:
			return WIPV4Address(address)
		except ValueError:
			pass

		try:
			return WFQDN(address)
		except ValueError:
			pass

		raise ValueError('Unable to parse address string. It must be an IP address or FQDN')

	@classmethod
	@verify_type(info=str)
	@verify_value(info=lambda x: len(x) > 0)
	def parse_socket_info(cls, info):
		""" Parse string that is formed like '[address]<:port>' and return corresponding
		:class:`.WIPV4ScketInfo` object

		:param info: string to parse

		:return: WIPV4ScketInfo
		"""
		info = info.split(':')
		if len(info) > 2:
			raise ValueError('Incorrect socket info specified')
		address = info[0].strip()
		port = int(info[1].strip()) if len(info) == 2 else None
		return WIPV4SocketInfo(address=address, port=port)



# -*- coding: utf-8 -*-
# wasp_general/template.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from abc import ABCMeta, abstractmethod
import io
from mako.template import Template
from mako.lookup import TemplateCollection
from mako.runtime import Context

from wasp_general.verify import verify_type, verify_value


class WTemplate(metaclass=ABCMeta):

	@abstractmethod
	def template(self):
		raise NotImplementedError('This method is abstract')


class WMakoTemplate(WTemplate):

	@verify_type(template=Template)
	def __init__(self, template):
		self.__template = template

	def template(self):
		return self.__template


class WTemplateText(WTemplate):

	@verify_type(text_template=str)
	def __init__(self, text_template, **kwargs):
		WTemplate.__init__(self)
		self.__template = Template(text=text_template, **kwargs)

	def template(self):
		return self.__template


class WTemplateFile(WTemplate):

	@verify_type(template_filename=str)
	@verify_value(template_filename=lambda x: len(x) > 0)
	def __init__(self, template_filename, **kwargs):
		WTemplate.__init__(self)
		self.__template = Template(filename=template_filename, **kwargs)

	def template(self):
		return self.__template


class WTemplateLookup(WTemplate):

	@verify_type(template_id=str, template_collection=TemplateCollection)
	def __init__(self, template_id, template_collection):
		WTemplate.__init__(self)
		self.__template_id = template_id
		self.__collection = template_collection

	def template(self):
		return self.__collection.get_template(self.__template_id)


class WTemplateRenderer:

	@verify_type(template=WTemplate, context=(None, dict))
	def __init__(self, template, context=None):
		self.__template = template
		self.__context = context if context is not None else {}

	def template(self):
		return self.__template.template()

	def context(self):
		return self.__context

	def update_context(self, **context_vars):
		self.__context.update(context_vars)

	def render(self, **context_vars):
		template = self.template()
		buffer = io.StringIO()
		context = self.context()
		context.update(context_vars)
		context = Context(buffer, **context)
		template.render_context(context)
		return buffer.getvalue()



# -*- coding: utf-8 -*-
# wasp_general/network/messenger/transport.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

from wasp_general.verify import verify_type

from wasp_general.network.messenger.proto import WMessengerOnionLayerProto, WMessengerOnionSessionProto
from wasp_general.network.messenger.proto import WMessengerEnvelopeProto
from wasp_general.network.messenger.envelope import WMessengerBytesEnvelope

from wasp_general.network.service import WZMQHandler, WZMQSyncAgent


class WMessengerSendAgentLayer(WMessengerOnionLayerProto):

	__layer_name__ = "com.binblob.wasp-general.send-agent-layer"
	""" Layer name
	"""

	def __init__(self):
		WMessengerOnionLayerProto.__init__(self, WMessengerSendAgentLayer.__layer_name__)

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerBytesEnvelope, send_agent=WZMQHandler.SendAgent, handler=WZMQHandler)
	def process(self, envelope, session, send_agent=None, handler=None, **kwargs):
		send_agent.send(handler, envelope.message())
		return envelope


class WMessengerSyncReceiveAgentLayer(WMessengerOnionLayerProto):
	__layer_name__ = "com.binblob.wasp-general.sync-receive-agent-layer"
	""" Layer name
	"""

	def __init__(self):
		WMessengerOnionLayerProto.__init__(self, WMessengerSyncReceiveAgentLayer.__layer_name__)

	@verify_type('paranoid', session=WMessengerOnionSessionProto)
	@verify_type(envelope=WMessengerEnvelopeProto, receive_agent=WZMQSyncAgent, timeout=(int, float, None))
	@verify_type(copy_meta=(bool, None))
	def process(self, envelope, session, receive_agent=None, timeout=None, copy_meta=None, **kwargs):
		timeout = timeout if timeout is not None else receive_agent.timeout()
		if receive_agent.event().wait(timeout) is not True:
			raise TimeoutError('Response timeout')
		meta = None
		if copy_meta is not None and copy_meta is True:
			meta = envelope
		return WMessengerBytesEnvelope(data=b''.join(receive_agent.data()), meta=meta)



# -*- coding: utf-8 -*-
# wasp_general/csv.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import csv

from wasp_general.verify import verify_type, verify_value


class WCSVExporter:

	@verify_type(titles=bool)
	def __init__(self, output_obj, titles=True, **csv_fmtparams):
		self.__output_obj = output_obj
		self.__titles = titles
		self.__csv_fmtparams = csv_fmtparams
		self.__csv_writer = None
		self.__omitted_fields = set()

	def output_obj(self):
		return self.__output_obj

	def titles(self):
		return self.__titles

	def csv_fmtparams(self):
		return self.__csv_fmtparams

	@verify_type(field_name=str)
	@verify_value(field_name=lambda x: len(x) > 0)
	def omit_field(self, field_name):
		self.__omitted_fields.add(field_name)

	def omitted_fields(self):
		return tuple(self.__omitted_fields)

	@verify_type(dict_record=dict)
	def export(self, dict_record):
		dict_record = self.__filter_field(dict_record)
		if self.__csv_writer is None:
			self.export_titles(dict_record)
		self.__check_record(dict_record)
		self.__csv_writer.writerow(dict_record)

	@verify_type(dict_record=dict)
	def export_titles(self, dict_record):
		if self.__csv_writer is not None:
			raise RuntimeError('Unable to export titles multiple time')
		fields = dict_record.keys()
		self.__csv_writer = csv.DictWriter(self.output_obj(), fieldnames=fields, **self.csv_fmtparams())
		if self.titles() is True:
			self.__csv_writer.writeheader()

	@verify_type(dict_record=dict)
	def __filter_field(self, dict_record):
		result = dict_record.copy()
		for field in self.omitted_fields():
			if field in result.keys():
				result.pop(field)
		return result

	@verify_type(dict_record=dict)
	def __check_record(self, dict_record):
		for key, value in dict_record.items():
			if isinstance(key, str) is False:
				raise TypeError('Invalid field name')
			if value is not None and isinstance(value, (str, int, float)) is False:
				raise TypeError('Invalid value for field "%s"' % key)



# -*- coding: utf-8 -*-
# wasp_general/network/clients/file.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import io
import os

from wasp_general.uri import WSchemeSpecification
from wasp_general.network.clients.base import WBasicNetworkClientProto
from wasp_general.network.clients.base import WBasicNetworkClientListDirCapability
from wasp_general.network.clients.base import WBasicNetworkClientChangeDirCapability
from wasp_general.network.clients.base import WBasicNetworkClientMakeDirCapability
from wasp_general.network.clients.base import WBasicNetworkClientCurrentDirCapability
from wasp_general.network.clients.base import WBasicNetworkClientUploadFileCapability
from wasp_general.network.clients.base import WBasicNetworkClientRemoveFileCapability


class WLocalFile(WBasicNetworkClientProto):

	def __init__(self, uri):
		WBasicNetworkClientProto.__init__(self, uri)
		self.__session_path = uri.path() if uri.path() is not None else '/'

	def session_path(self, value=None):
		if value is not None:
			self.__session_path = value
		return self.__session_path

	def _close(self):
		self.__session_path = None

	@classmethod
	def scheme_specification(cls):
		return WSchemeSpecification('file', path=WSchemeSpecification.ComponentDescriptor.required)

	@classmethod
	def agent_capabilities(cls):
		return WLocalFileListDirCapability, \
			WLocalFileMakeDirCapability, \
			WLocalFileChangeDirCapability, \
			WLocalFileUploadFileCapability, \
			WLocalFileRemoveFileCapability


class WLocalFileChangeDirCapability(WBasicNetworkClientChangeDirCapability):

	def request(self, path, *args, exec_cmd=None, **kwargs):
		if os.path.isdir(path) is True:
			self.network_agent().session_path(path)
			return True
		return False


class WLocalFileCurrentDirCapability(WBasicNetworkClientCurrentDirCapability):

	def request(self, *args, **kwargs):
		return self.network_agent().session_path()


class WLocalFileListDirCapability(WBasicNetworkClientListDirCapability):

	def request(self, *args, **kwargs):
		return tuple(os.listdir(self.network_agent().session_path()))


class WLocalFileMakeDirCapability(WBasicNetworkClientMakeDirCapability):

	def request(self, directory_name, *args, **kwargs):
		os.mkdir(os.path.join(self.network_agent().session_path(), directory_name))
		return True


class WLocalFileUploadFileCapability(WBasicNetworkClientUploadFileCapability):

	def request(self, file_name, file_obj, *args, **kwargs):
		with open(os.path.join(self.network_agent().session_path(), file_name), mode='wb') as f:
			chunk = file_obj.read(io.DEFAULT_BUFFER_SIZE)
			while len(chunk) > 0:
				f.write(chunk)
				chunk = file_obj.read(io.DEFAULT_BUFFER_SIZE)

		return True


class WLocalFileRemoveFileCapability(WBasicNetworkClientRemoveFileCapability):

	def request(self, file_name, *args, **kwargs):
		os.unlink(os.path.join(self.network_agent().session_path(), file_name))
		return True



# -*- coding: utf-8 -*-
# wasp_general/command/enhanced.py
#
# Copyright (C) 2017 the wasp-general authors and contributors
# <see AUTHORS file>
#
# This file is part of wasp-general.
#
# Wasp-general is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Wasp-general is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with wasp-general.  If not, see <http://www.gnu.org/licenses/>.

# TODO: document the code
# TODO: write tests for the code

# noinspection PyUnresolvedReferences
from wasp_general.version import __author__, __version__, __credits__, __license__, __copyright__, __email__
# noinspection PyUnresolvedReferences
from wasp_general.version import __status__

import re
from abc import abstractmethod
from enum import Enum

from wasp_general.verify import verify_type, verify_value, verify_subclass
from wasp_general.command.command import WCommandProto


class WCommandArgumentParsingError(Exception):
	pass


class WCommandArgumentDescriptor:

	class ArgumentCastingHelperProto:

		@abstractmethod
		@verify_type(argument_name=str)
		@verify_value(argument_name=lambda x: len(x) > 0)
		def cast(self, argument_name, argument_value):
			raise NotImplementedError('This method is abstract')

	class FlagArgumentCastingHelper(ArgumentCastingHelperProto):

		@verify_type(argument_name=str, argument_value=bool)
		@verify_value(argument_name=lambda x: len(x) > 0)
		def cast(self, argument_name, argument_value):
			return argument_value

	class ArgumentCastingHelper(ArgumentCastingHelperProto):

		@verify_type(error_message=(str, None))
		@verify_value(casting_fn=lambda x: x is None or callable(x))
		@verify_value(validate_fn=lambda x: x is None or callable(x))
		def __init__(self, casting_fn=None, validate_fn=None, error_message=None):
			WCommandArgumentDescriptor.ArgumentCastingHelperProto.__init__(self)
			self.__casting_fn = casting_fn
			self.__validate_fn = validate_fn
			self.__error_message = error_message

		def casting_function(self):
			return self.__casting_fn

		def validate_function(self):
			return self.__validate_fn

		def error_message(self):
			return self.__error_message

		@verify_type(argument_name=str, argument_value=str)
		@verify_value(argument_name=lambda x: len(x) > 0)
		def cast(self, argument_name, argument_value):
			casting_fn = self.casting_function()
			if casting_fn is not None:
				casted_argument_value = casting_fn(argument_value)
			else:
				casted_argument_value = argument_value
			validate_fn = self.validate_function()
			if validate_fn is not None:
				if validate_fn(casted_argument_value) is not True:
					error_message = self.error_message()
					if error_message is None:
						error_message = 'Attribute "%s" has invalid value: "%s"' % (
							argument_name, argument_value
						)
					raise WCommandArgumentParsingError(error_message)
			return casted_argument_value

	class StringArgumentCastingHelper(ArgumentCastingHelper):

		@verify_type('paranoid', error_message=(str, None))
		@verify_value('paranoid', validate_fn=lambda x: x is None or callable(x))
		def __init__(self, validate_fn=None, error_message=None):
			WCommandArgumentDescriptor.ArgumentCastingHelper.__init__(
				self, validate_fn=validate_fn, error_message=error_message
			)

	class IntegerArgumentCastingHelper(ArgumentCastingHelper):

		@verify_type('paranoid', error_message=(str, None))
		@verify_value('paranoid', validate_fn=lambda x: x is None or callable(x))
		@verify_type(base=int)
		@verify_value(base=lambda x: x > 0)
		def __init__(self, base=10, validate_fn=None, error_message=None):
			WCommandArgumentDescriptor.ArgumentCastingHelper.__init__(
				self, casting_fn=lambda x: int(x, base=base), validate_fn=validate_fn,
				error_message=error_message
			)

	class FloatArgumentCastingHelper(ArgumentCastingHelper):

		@verify_type('paranoid', error_message=(str, None))
		@verify_value('paranoid', validate_fn=lambda x: x is None or callable(x))
		def __init__(self, validate_fn=None, error_message=None):
			WCommandArgumentDescriptor.ArgumentCastingHelper.__init__(
				self, casting_fn=lambda x: float(x), validate_fn=validate_fn,
				error_message=error_message
			)

	class DataSizeArgumentHelper(ArgumentCastingHelper):

		__write_rate_re__ = re.compile('^(\d+[.\d]*)([KMGT]?)$')

		def __init__(self):
			WCommandArgumentDescriptor.ArgumentCastingHelper.__init__(
				self, casting_fn=self.cast_string
			)

		@staticmethod
		@verify_type(value=str)
		def cast_string(value):
			re_rate = WCommandArgumentDescriptor.DataSizeArgumentHelper.__write_rate_re__.search(value)
			if re_rate is None:
				raise ValueError('Invalid write rate')
			result = float(re_rate.group(1))
			if re_rate.group(2) == 'K':
				result *= (1 << 10)
			elif re_rate.group(2) == 'M':
				result *= (1 << 20)
			elif re_rate.group(2) == 'G':
				result *= (1 << 30)
			elif re_rate.group(2) == 'T':
				result *= (1 << 40)

			return result

	class EnumArgumentHelper(ArgumentCastingHelper):

		@verify_subclass(enum_cls=Enum)
		def __init__(self, enum_cls):
			WCommandArgumentDescriptor.ArgumentCastingHelper.__init__(
				self, casting_fn=self.cast_string
			)
			for item in enum_cls:
				if isinstance(item.value, str) is False:
					raise TypeError('Enum fields must bt str type')
			self.__enum_cls = enum_cls

		@verify_type(value=str)
		def cast_string(self, value):
			return self.__enum_cls(value)

	class RegExpArgumentHelper(ArgumentCastingHelper):

		@verify_type(regexp=str)
		def __init__(self, regexp):
			WCommandArgumentDescriptor.ArgumentCastingHelper.__init__(
				self, casting_fn=self.cast_string
			)
			self.__regexp = re.compile(regexp)

		def re(self):
			return self.__regexp

		def validate_function(self):
			return lambda x: x is not None

		@verify_type(value=str)
		def cast_string(self, value):
			result = self.re().match(value)
			if result is not None:
				return result.groups()

	@verify_type(argument_name=str, required=bool, flag_mode=bool, multiple_values=bool, help_info=(str, None))
	@verify_type(meta_var=(str, None), default_value=(str, None))
	@verify_value(argument_name=lambda x: len(x) > 0)
	def __init__(
		self, argument_name, required=False, flag_mode=False, multiple_values=False, help_info=None,
		meta_var=None, default_value=None, casting_helper=None
	):
		"""
		note: 'required' is useless for flag-mode attribute
		"""
		if (flag_mode is True and multiple_values is True) or \
			(flag_mode is True and default_value is not None) or \
			(multiple_values is True and default_value is not None):
				raise ValueError(
					'Argument has conflict options. "flag_mode" and "multiple_values" can not be '
					'used at the same time'
				)

		if casting_helper is not None:
			flag_helper = WCommandArgumentDescriptor.FlagArgumentCastingHelper
			general_helper = WCommandArgumentDescriptor.ArgumentCastingHelper

			if flag_mode is True and isinstance(casting_helper, flag_helper) is False:
				raise TypeError(
					'casting_helper must be an instance of '
					'WCommandArgumentDescriptor.FlagArgumentCastingHelper for flag-mode attribute'
				)
			elif flag_mode is False and isinstance(casting_helper, general_helper) is False:
				raise TypeError(
					'casting_helper must be an instance of '
					'WCommandArgumentDescriptor.ArgumentCastingHelper for every attribute except '
					'flag-mode attribute'
				)

		self.__argument_name = argument_name
		self.__flag_mode = flag_mode
		self.__multiple_values = multiple_values
		self.__default_value = default_value

		self.__required = required
		self.__help_info = help_info
		self.__meta_var = meta_var

		if casting_helper is not None:
			self.__casting_helper = casting_helper
		elif flag_mode is True:
			self.__casting_helper = WCommandArgumentDescriptor.FlagArgumentCastingHelper()
		else:
			self.__casting_helper = WCommandArgumentDescriptor.StringArgumentCastingHelper()

	def argument_name(self):
		return self.__argument_name

	def flag_mode(self):
		return self.__flag_mode

	def multiple_values(self):
		return self.__multiple_values

	def required(self):
		return self.__required

	def default_value(self):
		return self.__default_value

	def casting_helper(self):
		return self.__casting_helper

	@verify_type(argument_name=str)
	@verify_value(argument_name=lambda x: len(x) > 0)
	def cast(self, argument_name, argument_value):
		return self.casting_helper().cast(argument_name, argument_value)

	def help_info(self):
		return self.__help_info

	def meta_var(self):
		return self.__meta_var


class WCommandArgumentRelationship:

	class Relationship(Enum):
		conflict = 1
		requirement = 2
		one_of = 3

	@verify_type(argument_names=str)
	def __init__(self, relationship, *argument_names):
		if isinstance(relationship, WCommandArgumentRelationship.Relationship) is False:
			raise TypeError('Invalid relationship type')
		self.__relationship = relationship
		if len(argument_names) < 2:
			raise ValueError('Relationship can be made with 2 arguments and more')
		self.__arguments = argument_names

	def relationship(self):
		return self.__relationship

	def arguments(self):
		return self.__arguments


class WCommandArgumentParser:

	@verify_type(argument_descriptors=WCommandArgumentDescriptor, relationships=(list, tuple, set, None))
	def __init__(self, *argument_descriptors, relationships=None):
		self.__descriptors = argument_descriptors
		self.__relationships = relationships
		if self.__relationships is not None:
			for relation in self.__relationships:
				if isinstance(relation, WCommandArgumentRelationship) is False:
					raise TypeError('Invalid relationship type')
				for argument_name in relation.arguments():
					argument_found = False
					for descriptor in self.__descriptors:
						if argument_name == descriptor.argument_name():
							argument_found = True
							break
					if argument_found is False:
						raise ValueError('Relationship with unknown argument was specified')

	def descriptors(self):
		return self.__descriptors

	def relationships(self):
		return self.__relationships

	def __select_arguments(self, relation, parsed_result):
		result = []
		arguments = relation.arguments()
		for argument_name in arguments:
			if argument_name in parsed_result.keys():
				result.append(argument_name)
		return result

	def __check_conflict_relation(self, relation, parsed_result):
		conflict_arguments = self.__select_arguments(relation, parsed_result)
		if len(conflict_arguments) >= 1:
			raise WCommandArgumentParsingError(
				"Conflict arguments was found: %s" % (', '.join(conflict_arguments))
			)

	def __check_requirements_relation(self, relation, parsed_result):
		found_arguments = set(self.__select_arguments(relation, parsed_result))
		required_arguments = set(relation.arguments())

		not_found_arguments = required_arguments.difference(found_arguments)
		if len(not_found_arguments) > 0 and len(not_found_arguments) != len(required_arguments):
			raise WCommandArgumentParsingError(
				"Required arguments wasn't found: %s" % (', '.join(not_found_arguments))
			)

	def __check_one_of_relation(self, relation, parsed_result):
		found_arguments = set(self.__select_arguments(relation, parsed_result))

		if len(found_arguments) > 1:
			raise WCommandArgumentParsingError(
				"Conflict arguments was found: %s" % (', '.join(found_arguments))
			)
		elif len(found_arguments) == 0:
			arguments = relation.arguments()
			raise WCommandArgumentParsingError(
				"Required arguments was not found. It should be one of: %s" % (', '.join(arguments))
			)

	@verify_type(command_tokens=str)
	def parse(self, *command_tokens):
		descriptors = list(self.descriptors())
		command_tokens = list(command_tokens)
		result = {}
		while len(command_tokens) > 0:
			reduced_command_tokens, descriptors, next_result = \
				self.reduce_tokens(command_tokens.copy(), descriptors.copy(), previous_result=result)
			if len(reduced_command_tokens) >= len(command_tokens):
				raise WCommandArgumentParsingError("Command tokens wasn't reduce")

			command_tokens = reduced_command_tokens
			result = next_result

		relationships = self.relationships()
		if relationships is not None:
			for relation in relationships:
				relation_type = relation.relationship()
				if relation_type == WCommandArgumentRelationship.Relationship.conflict:
					self.__check_conflict_relation(relation, result)
				elif relation_type == WCommandArgumentRelationship.Relationship.requirement:
					self.__check_requirements_relation(relation, result)
				elif relation_type == WCommandArgumentRelationship.Relationship.one_of:
					self.__check_one_of_relation(relation, result)
				else:
					raise RuntimeError('Unknown relationship was specified')

		for descriptor in descriptors:
			argument_name = descriptor.argument_name()
			if descriptor.flag_mode() is True:
				result[argument_name] = descriptor.cast(argument_name, False)
			if descriptor.default_value() is not None:
				result[argument_name] = descriptor.cast(argument_name, descriptor.default_value())

		for descriptor in self.descriptors():
			if descriptor.required() is True:
				argument_name = descriptor.argument_name()
				if argument_name not in result.keys():
					raise WCommandArgumentParsingError(
						"Required argument wasn't found: %s" % argument_name
					)

		return result

	@classmethod
	@verify_type(command_tokens=list, argument_descriptors=list, previous_result=(dict, None))
	def reduce_tokens(cls, command_tokens, argument_descriptors, previous_result=None):
		argument_name = command_tokens.pop(0)

		descriptor = None
		for i in range(len(argument_descriptors)):
			descriptor_to_check = argument_descriptors[i]
			if descriptor_to_check.argument_name() == argument_name:
				descriptor = descriptor_to_check
				if descriptor_to_check.multiple_values() is False:
					argument_descriptors.pop(i)
				break
		if descriptor is None:
			if argument_name in previous_result.keys():
				raise WCommandArgumentParsingError(
					'Multiple argument ("%s") values found' % argument_name
				)
			else:
				raise WCommandArgumentParsingError('Unknown argument: "%s"' % argument_name)

		result = previous_result.copy() if previous_result is not None else {}
		if descriptor.flag_mode() is True:
			result[argument_name] = descriptor.cast(argument_name, True)
		else:
			if len(command_tokens) == 0:
				raise WCommandArgumentParsingError("Argument requires value. Value wasn't found")
			argument_value = descriptor.cast(argument_name, command_tokens.pop(0))

			if descriptor.multiple_values() is True:
				if argument_name not in result.keys():
					result[argument_name] = [argument_value]
				else:
					result[argument_name].append(argument_value)
			else:
				if argument_name not in result.keys():
					result[argument_name] = argument_value
				else:
					raise WCommandArgumentParsingError(
						'Multiple values spotted for the single argument'
					)

		return command_tokens, argument_descriptors, result

	def arguments_help(self):
		result = []
		for argument in self.descriptors():
			argument_name = argument.argument_name()
			if argument.flag_mode() is not True:
				value_name = argument.meta_var()
				if value_name is None:
					value_name = 'value'
				argument_name = '%s [%s]' % (argument_name, value_name)

			description = argument.help_info()
			if description is None:
				description = 'argument description unavailable'

			meta = []
			if argument.required() is True:
				meta.append('required')

			if argument.multiple_values() is True:
				meta.append('may have multiple values')

			default_value = argument.default_value()
			if default_value is not None:
				meta.append('default value: %s' % default_value)

			if len(meta) > 0:
				description += (' (%s)' % ', '.join(meta))

			result.append((argument_name, description))
		return tuple(result)


class WEnhancedCommand(WCommandProto):

	@verify_type('paranoid', argument_descriptors=WCommandArgumentDescriptor, relationships=(list, tuple, set, None))
	@verify_type(command=str)
	@verify_value(command=lambda x: len(x) > 0)
	def __init__(self, command, *argument_descriptors, relationships=None):
		WCommandProto.__init__(self)
		self.__command = command
		self.__arguments_descriptors = argument_descriptors
		self.__relationships = relationships
		self.__parser = WCommandArgumentParser(*self.argument_descriptors(), relationships=self.relationships())

	def command_token(self):
		return self.__command

	def argument_descriptors(self):
		return self.__arguments_descriptors

	def relationships(self):
		return self.__relationships

	def parser(self):
		return self.__parser

	@verify_type(command_tokens=str)
	def match(self, *command_tokens, **command_env):
		if len(command_tokens) > 0:
			if command_tokens[0] == self.command_token():
				return True
		return False

	@verify_type(command_tokens=str)
	def exec(self, *command_tokens, **command_env):
		if len(command_tokens) > 0:
			if command_tokens[0] == self.command_token():
				return self._exec(self.parser().parse(*command_tokens[1:]), **command_env)
		raise RuntimeError('Invalid tokens')

	@abstractmethod
	@verify_type(command_arguments=dict)
	def _exec(self, command_arguments, **command_env):
		raise NotImplementedError('This method is abstract')

	def arguments_help(self):
		return self.parser().arguments_help()

	def command_help(self):
		arguments_help = self.arguments_help()
		if len(arguments_help) == 0:
			return 'Command does not have arguments\n'

		info = 'Command arguments:\n'
		for argument_name, argument_description in arguments_help:
			info += '\t%s - %s\n' % (argument_name, argument_description)
		return info



# -*- coding: utf-8 -*-

import pytest

from http.cookies import SimpleCookie

from wasp_general.network.web.cookies import WHTTPCookie, WHTTPCookieJar


class TestWHTTPCookies:

	def test_cookie(self):
		cookie = WHTTPCookie('cookie1', 'value1', path='/---')
		assert(isinstance(cookie, WHTTPCookie) is True)
		assert(cookie.name() == 'cookie1')
		assert(cookie.value() == 'value1')
		assert(cookie.value('value2') == 'value2')
		assert(cookie.attr('path') == '/---')
		assert(cookie.attr('path', '/!!!') == '/!!!')
		assert(cookie.attrs_as_dict() == {'path': '/!!!'})
		pytest.raises(ValueError, cookie.attr, 'path', ';;;')

		assert(str(cookie) == 'Set-Cookie: cookie1=value2; Path=/!!!')

		copy_cookie = cookie.copy()
		assert(str(copy_cookie) == 'Set-Cookie: cookie1=value2; Path=/!!!')
		copy_cookie.remove_attr('path')
		assert(str(copy_cookie) == 'Set-Cookie: cookie1=value2')

		ro_cookie = cookie.ro()
		assert(str(ro_cookie) == 'Set-Cookie: cookie1=value2; Path=/!!!')
		pytest.raises(RuntimeError, ro_cookie.value, 'zzz')
		pytest.raises(RuntimeError, ro_cookie.attr, 'path', '/zzz')
		pytest.raises(RuntimeError, ro_cookie.remove_attr, 'path')

		cookie = WHTTPCookie('cookie2', 'value0', max_age='1/1/1')
		assert(str(cookie) == 'Set-Cookie: cookie2=value0; Max-Age=1/1/1')
		cookie.attr('max-age', '2/2/2')
		assert(str(cookie) == 'Set-Cookie: cookie2=value0; Max-Age=2/2/2')
		pytest.raises(ValueError, cookie.attr, 'ZZZZ', '3/3/3')


class TestWHTTPCookieJar:

	def test_cookie_jar(self):
		cookie_jar = WHTTPCookieJar()
		assert(isinstance(cookie_jar, WHTTPCookieJar) is True)
		assert(cookie_jar.cookies() == tuple())

		cookie = WHTTPCookie('cookie1', 'value1')
		cookie_jar.add_cookie(cookie)
		assert(cookie_jar.cookies() == ('cookie1',))
		assert(cookie_jar['cookie1'] == cookie)
		assert(cookie_jar['cookie1'].value() == 'value1')
		cookie.value('value2')
		assert(cookie_jar['cookie1'].value() == 'value2')
		assert([x for x in cookie_jar] == [cookie])

		ro_cookie_jar = cookie_jar.ro()
		pytest.raises(RuntimeError, ro_cookie_jar.remove_cookie, 'cookie1')
		pytest.raises(RuntimeError, ro_cookie_jar.add_cookie, WHTTPCookie('cookie2', 'value2'))
		pytest.raises(RuntimeError, ro_cookie_jar['cookie1'].value, 'value3')

		cookie_jar.remove_cookie('cookie1')
		assert(cookie_jar.cookies() == tuple())

	def test_import_simple_cookie(self):
		simple_cookie = SimpleCookie()
		simple_cookie['cookie1'] = 'value1'
		simple_cookie['cookie1']['path'] = '/'
		simple_cookie['cookie2'] = 'value2'
		simple_cookie['cookie2']['domain'] = 'foo.bar'

		cookie_jar = WHTTPCookieJar.import_simple_cookie(simple_cookie)
		cookies = list(cookie_jar.cookies())
		cookies.sort()
		assert(cookies == ['cookie1', 'cookie2'])
		assert(cookie_jar['cookie1'].value() == 'value1')
		assert(cookie_jar['cookie2'].value() == 'value2')
		assert(str(cookie_jar['cookie1']) == 'Set-Cookie: cookie1=value1; Path=/')
		assert(str(cookie_jar['cookie2']) == 'Set-Cookie: cookie2=value2; Domain=foo.bar')

	def test_import_header_text(self):
		http_code = 'Set-Cookie: cookie1=value1; Path=/;, cookie2=value2; Domain=foo.bar\r\n'

		cookie_jar = WHTTPCookieJar.import_header_text(http_code)
		cookies = list(cookie_jar.cookies())
		cookies.sort()
		assert(cookies == ['cookie1', 'cookie2'])
		assert(cookie_jar['cookie1'].value() == 'value1')
		assert(cookie_jar['cookie2'].value() == 'value2')
		assert(str(cookie_jar['cookie1']) == 'Set-Cookie: cookie1=value1; Path=/')
		assert(str(cookie_jar['cookie2']) == 'Set-Cookie: cookie2=value2; Domain=foo.bar')



# -*- coding: utf-8 -*-

import pytest

from wasp_general.network.web.proto import WWebResponseProto, WWebRequestProto, WWebSessionProto, WWebPresenter
from wasp_general.network.web.proto import WWebTargetRouteProto, WWebRouteMapProto, WWebErrorPresenter
from wasp_general.network.web.proto import WWebServiceProto, WWebPresenterCollectionProto, WWebPresenterFactoryProto

from wasp_general.network.web.headers import WHTTPHeaders


def test_abstract():
	pytest.raises(TypeError, WWebRequestProto)
	pytest.raises(NotImplementedError, WWebRequestProto.session, None)
	pytest.raises(NotImplementedError, WWebRequestProto.method, None)
	pytest.raises(NotImplementedError, WWebRequestProto.path, None)
	pytest.raises(NotImplementedError, WWebRequestProto.headers, None)

	pytest.raises(TypeError, WWebSessionProto)
	pytest.raises(NotImplementedError, WWebSessionProto.client_address, None)
	pytest.raises(NotImplementedError, WWebSessionProto.server_address, None)
	pytest.raises(NotImplementedError, WWebSessionProto.protocol_version, None)
	pytest.raises(NotImplementedError, WWebSessionProto.protocol, None)
	pytest.raises(NotImplementedError, WWebSessionProto.read_request, None)
	pytest.raises(
		NotImplementedError, WWebSessionProto.write_response, None,
		TestWWebRequestProto.Request(), TestWWebResponseProto()
	)
	pytest.raises(NotImplementedError, WWebSessionProto.session_close, None)

	pytest.raises(TypeError, WWebTargetRouteProto)
	pytest.raises(NotImplementedError, WWebTargetRouteProto.presenter_name, None)
	pytest.raises(NotImplementedError, WWebTargetRouteProto.presenter_action, None)

	class Service(WWebServiceProto):

		def route_map(self):
			return

		def presenter_collection(self):
			return

		def presenter_factory(self):
			return

		def execute(self, request, target_route):
			return

	pytest.raises(TypeError, WWebRouteMapProto)
	pytest.raises(NotImplementedError, WWebRouteMapProto.route, None, TestWWebRequestProto.Request(), Service())
	pytest.raises(NotImplementedError, WWebRouteMapProto.error_presenter, None)

	pytest.raises(TypeError, WWebErrorPresenter)
	pytest.raises(NotImplementedError, WWebErrorPresenter.error_code, None, 100)
	pytest.raises(NotImplementedError, WWebErrorPresenter.exception_error, None, Exception(''))

	pytest.raises(TypeError, WWebServiceProto)
	pytest.raises(NotImplementedError, WWebServiceProto.route_map, None)
	pytest.raises(NotImplementedError, WWebServiceProto.presenter_collection, None)
	pytest.raises(NotImplementedError, WWebServiceProto.presenter_factory, None)
	pytest.raises(
		NotImplementedError, WWebServiceProto.execute, None,
		TestWWebRequestProto.Request(), TestWWebTargetRouteProto.Route()
	)

	pytest.raises(TypeError, WWebPresenterCollectionProto)
	pytest.raises(NotImplementedError, WWebPresenterCollectionProto.presenter, None, 'presenter')
	pytest.raises(NotImplementedError, WWebPresenterCollectionProto.has, None, 'presenter')

	pytest.raises(TypeError, WWebPresenter)
	pytest.raises(NotImplementedError, WWebPresenter.__presenter_name__)

	pytest.raises(TypeError, WWebPresenterFactoryProto)
	pytest.raises(NotImplementedError, WWebPresenterFactoryProto.instantiable, None, TestWWebPresenter.Presenter)
	pytest.raises(NotImplementedError, WWebPresenterFactoryProto.instantiate, None, TestWWebPresenter.Presenter)


class TestWWebResponseProto:

	def test_response(self):
		response = WWebResponseProto()
		assert(isinstance(response, WWebResponseProto) is True)
		assert(response.status() is None)
		assert(response.headers() is None)
		assert(response.response_data() is None)
		assert(response.__pushed_responses__() == tuple())


class TestWWebRequestProto:

	class Request(WWebRequestProto):
		def session(self):
			pass

		def method(self):
			return 'GET'

		def path(self):
			return '/foo/bar?test_var=value?&test_var=&foo=%02/bar#/zzz?'

		def headers(self):
			headers = WHTTPHeaders()
			headers.add_headers('Host', 'foo.bar:8080')
			headers.add_headers('Content-type', 'application/x-www-form-urlencoded')
			return headers

	def test_request(self):

		request = TestWWebRequestProto.Request()
		assert(request.request_data() is None)
		assert(request.virtual_host() == 'foo.bar:8080')
		assert(request.content_type() == 'application/x-www-form-urlencoded')

		get_vars = request.get_vars()
		assert(get_vars['foo'] == ['\x02/bar'])
		get_vars = get_vars['test_var']
		get_vars.sort()
		assert(get_vars == ['', 'value?'])
		request.method = lambda: 'POST'
		pytest.raises(RuntimeError, request.get_vars)

		request.request_data = lambda: b'test_var=post_value?&test_var=&zzz=/xxx%01#/zzz?'
		post_vars = request.post_vars()
		assert(post_vars['zzz'] == ['/xxx\x01'])
		post_vars = post_vars['test_var']
		post_vars.sort()
		post_vars.sort()
		assert(post_vars == ['', 'post_value?'])

		request.method = lambda: 'GET'
		pytest.raises(RuntimeError, request.post_vars)
		request.method = lambda: 'POST'
		request.headers = lambda: None
		pytest.raises(RuntimeError, request.post_vars)


class TestWWebPresenter:

	class Presenter(WWebPresenter):

		@classmethod
		def __presenter_name__(cls):
			return 'preseter name'

	def test_presenter(self):
		request = TestWWebRequestProto.Request()
		presenter = TestWWebPresenter.Presenter(request)

		assert(isinstance(presenter, WWebPresenter) is True)
		assert(presenter.__request__() == request)


class TestWWebTargetRouteProto:

	class DummyPresenter:

		zzz = '!'

		def __init__(self, request):
			pass

		def foo(self, a, b=5):
			return a * b

		def _bar(self):
			return 1

	class Route(WWebTargetRouteProto):

		def route_map(self):
			return None

		def presenter_name(self):
			return 'test presenter'

		def presenter_action(self):
			return 'foo'

	def test_route(self):
		route = TestWWebTargetRouteProto.Route()
		assert(route.presenter_args() == {})


class TestWWebRouteMapProto:

	class RouteMap(WWebRouteMapProto):

		def route(self, request):
			pass

		def error_presenter(self):
			pass

	def test_route(self):
		route_map = TestWWebRouteMapProto.RouteMap()

		target_route = TestWWebTargetRouteProto.Route()
		assert(target_route.presenter_action() == 'foo')
		assert(route_map.target_route_valid(target_route) is True)
		target_route.presenter_action = lambda: '_bar'
		assert(route_map.target_route_valid(target_route) is False)
